"""A text base user interface for euporie."""

from __future__ import annotations

import logging
from functools import partial
from typing import TYPE_CHECKING, cast
from weakref import WeakKeyDictionary

from prompt_toolkit.filters import Condition
from prompt_toolkit.formatted_text.base import to_formatted_text
from prompt_toolkit.layout.containers import (
    ConditionalContainer,
    DynamicContainer,
    Float,
    FloatContainer,
    WindowAlign,
)
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit.layout.dimension import Dimension

from euporie.core.app.app import BaseApp
from euporie.core.bars.command import CommandBar
from euporie.core.bars.menu import ToolbarCompletionsMenu
from euporie.core.bars.search import SearchBar
from euporie.core.bars.status import StatusBar, StatusContainer
from euporie.core.commands import get_cmd
from euporie.core.filters import has_tabs
from euporie.core.ft.utils import truncate
from euporie.core.key_binding.registry import register_bindings
from euporie.core.layout.containers import HSplit, VSplit, Window
from euporie.core.layout.decor import Pattern
from euporie.core.widgets.dialog import (
    AboutDialog,
    ConfirmDialog,
    ErrorDialog,
    MsgBoxDialog,
    NoKernelsDialog,
    OpenFileDialog,
    SaveAsDialog,
    SelectKernelDialog,
    ShortcutsDialog,
    UnsavedDialog,
)
from euporie.core.widgets.file_browser import FileBrowser
from euporie.core.widgets.layout import TabBarControl, TabBarTab
from euporie.core.widgets.logo import Logo
from euporie.core.widgets.menu import MenuBar, MenuItem
from euporie.core.widgets.pager import Pager
from euporie.core.widgets.palette import CommandPalette
from euporie.notebook.enums import TabMode
from euporie.notebook.widgets.side_bar import SideBar

if TYPE_CHECKING:
    from asyncio import AbstractEventLoop
    from collections.abc import Sequence
    from typing import Any, Callable

    from prompt_toolkit.application.application import Application
    from prompt_toolkit.formatted_text import StyleAndTextTuples
    from prompt_toolkit.layout.containers import AnyContainer

    from euporie.core.bars.status import StatusBarFields
    from euporie.core.tabs import TabRegistryEntry
    from euporie.core.tabs.base import Tab
    from euporie.core.widgets.cell import Cell
    from euporie.notebook.tabs.notebook import Notebook

log = logging.getLogger(__name__)


class NotebookApp(BaseApp):
    """Notebook app.

    Interactively edit a notebook file.

    Launch the interactive TUI notebook editor, allowing you to run and edit Jupyter
    notebooks in the terminal.
    """

    _tab_container: AnyContainer

    name = "notebook"

    def __init__(self, **kwargs: Any) -> None:
        """Create a new euporie text user interface application instance."""
        kwargs.setdefault("title", "euporie-notebook")
        kwargs.setdefault("full_screen", True)
        kwargs.setdefault("leave_graphics", False)
        super().__init__(**kwargs)
        self.bindings_to_load.append("euporie.notebook.app.NotebookApp")

        self._tab_bar_tabs: dict[int, WeakKeyDictionary[Tab, TabBarTab]] = {}
        self.on_tabs_change += self.set_tab_container

        # Register config hooks
        self.config.events.show_cell_borders += lambda x: self.refresh()
        self.config.events.tab_mode += self.set_tab_container
        self.config.events.background_pattern += self.set_tab_container
        self.config.events.background_character += self.set_tab_container

    def pre_run(self, app: Application | None = None) -> None:
        """Continue loading the app."""
        super().pre_run(app)
        # Add a toolbar completion menu
        self.menus["toolbar_completions"] = Float(
            content=ToolbarCompletionsMenu(), ycursor=True, transparent=True
        )

    @property
    def tab_registry(self) -> list[TabRegistryEntry]:
        """Return the tab registry."""
        from euporie.notebook.tabs import _TAB_REGISTRY

        return _TAB_REGISTRY

    def format_title(self) -> StyleAndTextTuples:
        """Format the tab's title for display in the top right of the app."""
        if self.tabs:
            # Get tab without re-focusing it
            tab = self.tabs[self._tab_idx]
            title = truncate(to_formatted_text(tab.title, style="bold"), 30)
            return [("", " "), *title, ("", " ")]
        else:
            return []

    def tab_container(self) -> AnyContainer:
        """Return a container with all opened tabs.

        Returns:
            A layout container displaying the opened tab containers.
        """
        try:
            return self._tab_container
        except AttributeError:
            self.set_tab_container()
            return self._tab_container

    def set_tab_container(self, app: BaseApp | None = None) -> None:
        """Set the container to use to display opened tabs."""
        tab_mode = TabMode(self.config.tab_mode)
        if not self.tabs:
            self._tab_container = Pattern(
                self.config.background_character,
                self.config.background_pattern,
            )
        elif tab_mode == TabMode.TILE_HORIZONTALLY:
            children = []
            for tab in self.tabs:

                def _get_tab_container(tab: Tab = tab) -> Tab:
                    return tab

                children.append(DynamicContainer(_get_tab_container))
            self._tab_container = HSplit(
                children=children,
                padding=1,
                padding_style="class:tab-padding",
                padding_char="─",
            )
        elif tab_mode == TabMode.TILE_VERTICALLY:
            children = []
            for tab in self.tabs:

                def _get_tab_container(tab: Tab = tab) -> Tab:
                    return tab

                children.append(DynamicContainer(_get_tab_container))
            self._tab_container = VSplit(
                children=children,
                padding=1,
                padding_style="class:tab-padding",
                padding_char="│",
            )
        else:
            self._tab_container = HSplit(
                [
                    ConditionalContainer(
                        Window(
                            TabBarControl(
                                tabs=self.tab_bar_tabs, active=lambda: self._tab_idx
                            ),
                            height=2,
                            style="class:app-tab-bar",
                            dont_extend_height=True,
                        ),
                        filter=Condition(
                            lambda: (
                                len(self.tabs) > 1 or self.config.always_show_tab_bar
                            )
                            and TabMode(self.config.tab_mode) == TabMode.STACK
                        ),
                    ),
                    DynamicContainer(lambda: self.tabs[self._tab_idx]),
                ]
            )

    def _statusbar_defaults(self) -> StatusBarFields | None:
        """Load the default statusbar fields (run after keybindings are loaded)."""
        return (
            [
                [
                    ("", "Press "),
                    ("bold", get_cmd("new-notebook").key_str()),
                    ("", " to start a new notebook"),
                ],
            ],
            [[("", "Press "), ("bold", get_cmd("quit").key_str()), ("", " to quit")]],
        )

    def load_container(self) -> FloatContainer:
        """Build the main application layout."""
        self.logo = StatusContainer(body=Logo(), status=self._statusbar_defaults)

        title_bar = ConditionalContainer(
            Window(
                content=FormattedTextControl(self.format_title, show_cursor=False),
                height=1,
                style="class:menu.item",
                dont_extend_width=True,
                align=WindowAlign.RIGHT,
            ),
            filter=has_tabs,
        )

        self.pager = Pager()
        self.search_bar = SearchBar()
        self.command_bar = CommandBar()

        self.dialogs["command-palette"] = CommandPalette(self)
        self.dialogs["about"] = AboutDialog(self)
        self.dialogs["open-file"] = OpenFileDialog(self)
        self.dialogs["save-as"] = SaveAsDialog(self)
        self.dialogs["no-kernels"] = NoKernelsDialog(self)
        self.dialogs["change-kernel"] = SelectKernelDialog(self)
        self.dialogs["confirm"] = ConfirmDialog(self)
        self.dialogs["error"] = ErrorDialog(self)
        self.dialogs["unsaved"] = UnsavedDialog(self)
        self.dialogs["shortcuts"] = ShortcutsDialog(self)
        self.dialogs["msgbox"] = MsgBoxDialog(self)

        top_bar = ConditionalContainer(
            content=VSplit(
                [
                    self.logo,
                    MenuBar(app=self, menu_items=self.load_menu_items()),
                    title_bar,
                ]
            ),
            filter=self.config.filters.show_top_bar,
        )

        titles: Sequence[str]
        icons: Sequence[str]
        panels: Sequence[AnyContainer]
        titles, icons, panels = list(
            zip(
                *[
                    (
                        "File Browser",
                        "",
                        FileBrowser(width=25, on_open=self.open_file),
                    )
                ]
            ),
        )
        self.side_bar = SideBar(titles, icons, panels)

        self.container = FloatContainer(
            content=HSplit(
                [
                    top_bar,
                    VSplit(
                        [
                            self.side_bar,
                            HSplit(
                                [
                                    DynamicContainer(self.tab_container),
                                    self.pager,
                                ],
                                width=Dimension(weight=1),
                            ),
                        ],
                        height=Dimension(min=1),
                    ),
                    self.command_bar,
                    self.search_bar,
                    StatusBar(),
                ],
                style="class:body",
            ),
            floats=cast("list[Float]", self.floats),
        )

        return self.container

    def tab_bar_tabs(self) -> list[TabBarTab]:
        """Return a list of the current tabs for the tab-bar."""
        result = []
        for i, tab in enumerate(self.tabs):
            index_dict = self._tab_bar_tabs.setdefault(i, WeakKeyDictionary())
            if tab not in index_dict:
                index_dict[tab] = TabBarTab(
                    title=lambda tab=tab: tab.title,  # type: ignore [misc]
                    on_activate=partial(setattr, self, "tab_idx", i),
                    on_close=partial(self.close_tab, tab),
                    closeable=True,
                )
            result.append(self._tab_bar_tabs[i][tab])
        return result

    def _handle_exception(
        self, loop: AbstractEventLoop, context: dict[str, Any]
    ) -> None:
        exception = context.get("exception")
        # Also display a dialog to the user
        self.dialogs["error"].show(exception=exception)
        # Log observed exceptions to the log
        log.exception("An unhandled exception occurred", exc_info=exception)

    def exit(self, *args: Any, **kwargs: Any) -> None:
        """Check for unsaved files before closing.

        Creates a chain of close file commands, where the callback for each triggers
        the closure of the next. The closing process can be cancelled anywhere along
        the chain.

        Args:
            args: Positional arguments
            kwargs: Key word arguments

        """
        really_close = super().exit
        if self.tabs:

            def final_cb() -> None:
                """Really exit after the last tab in the chain is closed."""
                self.cleanup_closed_tab(self.tabs[0])
                really_close(*args, **kwargs)

            def create_cb(close_tab: Tab, cleanup_tab: Tab, cb: Callable) -> Callable:
                """Generate a tab close chaining callbacks.

                Cleans up after the previously closed tab, and requests to close the
                next tab in the chain.

                Args:
                    close_tab: The tab to close
                    cleanup_tab: The previously closed tab to cleanup
                    cb: The callback to call when work is complete

                Returns:
                    A callback function which cleans up `cleanup_tab` and closes
                        `close_tab`.

                """

                def inner() -> None:
                    self.cleanup_closed_tab(cleanup_tab)
                    close_tab.close(cb=cb)

                return inner

            cb = final_cb
            for close_tab, cleanup_tab in zip(self.tabs, self.tabs[1:]):
                cb = create_cb(close_tab, cleanup_tab, cb)
            self.tabs[-1].close(cb)
        else:
            really_close(*args, **kwargs)

    @property
    def notebook(self) -> Notebook | None:
        """Return the currently active notebook."""
        from euporie.notebook.tabs.notebook import Notebook

        if isinstance(self.tab, Notebook):
            return self.tab
        return None

    @property
    def cell(self) -> Cell | None:
        """Return the currently active cell."""
        from euporie.notebook.tabs.notebook import Notebook

        if isinstance(self.tab, Notebook):
            return self.tab.cell
        return None

    def load_menu_items(self) -> list[MenuItem]:
        """Load the list of menu items to display in the menu."""
        separator = MenuItem(separator=True)
        return [
            MenuItem(
                "File",
                children=[
                    get_cmd("new-notebook").menu,
                    get_cmd("open-file").menu,
                    separator,
                    get_cmd("save-file").menu,
                    get_cmd("save-as").menu,
                    get_cmd("close-tab").menu,
                    separator,
                    get_cmd("quit").menu,
                ],
                description="File management",
            ),
            MenuItem(
                "Edit",
                children=[
                    get_cmd("undelete-cells").menu,
                    separator,
                    get_cmd("cut-cells").menu,
                    get_cmd("copy-cells").menu,
                    get_cmd("paste-cells").menu,
                    separator,
                    get_cmd("copy-outputs").menu,
                    get_cmd("clear-cell-outputs").menu,
                    separator,
                    get_cmd("find").menu,
                    separator,
                    get_cmd("reformat-cells").menu,
                    get_cmd("reformat-notebook").menu,
                ],
                description="Make changes to the current file",
            ),
            MenuItem(
                "Run",
                children=[
                    get_cmd("run-selected-cells").menu,
                    get_cmd("run-all-cells").menu,
                ],
                description="Run notebook cells",
            ),
            MenuItem(
                "Kernel",
                children=[
                    get_cmd("interrupt-kernel").menu,
                    get_cmd("restart-kernel").menu,
                    get_cmd("change-kernel").menu,
                ],
                description="Manage the current notebook's kernel",
            ),
            MenuItem(
                "Tabs",
                children=[
                    get_cmd("next-tab").menu,
                    get_cmd("previous-tab").menu,
                    separator,
                    MenuItem(
                        "Tab mode",
                        children=[
                            get_cmd(f"set-tab-mode-{choice}").menu
                            for choice in self.config.choices.tab_mode
                        ],
                    ),
                ],
                description="Tab management",
            ),
            MenuItem(
                "Settings",
                children=[
                    MenuItem(
                        "UI Elements",
                        children=[
                            get_cmd("switch-background-pattern").menu,
                            get_cmd("toggle-show-cell-borders").menu,
                            get_cmd("toggle-always-show-tab-bar").menu,
                            get_cmd("toggle-show-side-bar").menu,
                            get_cmd("toggle-show-status-bar").menu,
                            get_cmd("toggle-show-scroll-bar").menu,
                            get_cmd("toggle-multiplexer-passthrough").menu,
                        ],
                        description="Turn elements of euporie's interface on or off",
                    ),
                    self.config.menus.color_scheme,
                    MenuItem(
                        "Syntax highlighting",
                        children=[
                            self.config.menus.syntax_highlighting,
                            separator,
                            *self.config.menus.syntax_theme.children,
                        ],
                        description="Configure syntax highlighting",
                    ),
                    get_cmd("toggle-expand").menu,
                    get_cmd("toggle-line-numbers").menu,
                    self.config.menus.set_cursor_shape,
                    separator,
                    self.config.menus.edit_mode,
                    separator,
                    MenuItem(
                        "Code tools",
                        children=[
                            get_cmd("toggle-enable-language-servers").menu,
                            separator,
                            self.config.menus.autosuggest,
                            get_cmd("toggle-autoformat").menu,
                            get_cmd("toggle-autocomplete").menu,
                            get_cmd("toggle-autoinspect").menu,
                        ],
                        description="Turn code assistance tools on or off",
                    ),
                    get_cmd("toggle-run-after-external-edit").menu,
                ],
                description="Make changes to euporie's configuration",
            ),
            MenuItem(
                "Help",
                children=[
                    get_cmd("show-command-palette").menu,
                    get_cmd("keyboard-shortcuts").menu,
                    get_cmd("view-documentation").menu,
                    separator,
                    get_cmd("view-logs").menu,
                    separator,
                    get_cmd("about").menu,
                ],
                description="Get help",
            ),
        ]

    # ################################# Key Bindings ##################################

    register_bindings(
        {
            "euporie.notebook.app.NotebookApp": {
                "new-notebook": "c-n",
                "toggle-show-top-bar": ("A-m"),
            }
        }
    )



"""Contain a registry of Comm target classes."""

from __future__ import annotations

from typing import TYPE_CHECKING

from euporie.core.comm.base import UnimplementedComm
from euporie.core.comm.ipywidgets import open_comm_ipywidgets

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import Any, Callable

    from euporie.core.comm.base import Comm, KernelTab

TARGET_CLASSES: dict[str, Callable[[KernelTab, str, dict, Sequence[bytes]], Comm]] = {
    "jupyter.widget": open_comm_ipywidgets
}


def open_comm(
    comm_container: KernelTab,
    content: dict[str, Any],
    buffers: Sequence[bytes],
) -> Comm:
    """Create a new object respsenting a Comm.

    The class used to represent the Comm is determined by the "target_class" given in
    the ``comm_open`` message.

    Args:
        comm_container: The notebook this comm belongs to
        content: The content of the ``comm_open`` message
        buffers: A list of binary data buffers sent with the ``comm_open`` message

    Returns:
        A class representing the comm

    """
    target_name = content.get("target_name", "")
    return TARGET_CLASSES.get(target_name, UnimplementedComm)(
        # comm_container=
        comm_container,
        # comm_id=
        str(content.get("comm_id")),
        # data=
        content.get("data", {}),
        # buffers=
        buffers,
    )



"""Settings for the console app."""

from __future__ import annotations

from typing import TYPE_CHECKING

from upath import UPath

from euporie.core.config import add_setting

if TYPE_CHECKING:
    from typing import TypeVar

    _AppResult = TypeVar("_AppResult")

add_setting(
    name="mouse_support",
    group="euporie.console.app",
    flags=["--mouse-support"],
    type_=bool,
    help_="Enable or disable mouse support",
    default=None,
    description="""
            When set to True, mouse support is enabled. When set to False, mouse
            support is disabled.
        """,
)
add_setting(
    name="max_stored_outputs",
    group="euporie.console.app",
    flags=["--max-stored-outputs"],
    type_=int,
    help_="The number of inputs / outputs to store in an in-memory notebook",
    default=100,
    schema={
        "minimum": 0,
    },
    description="""
        Defines the maximum number of executed "cells" to store in case the console
        session is saved to a file or converted into a notebook.
    """,
)

add_setting(
    name="connection_file",
    group="euporie.console.app",
    flags=["--connection-file", "--kernel-connection-file"],
    type_=UPath,
    help_="Attempt to connect to an existing kernel using a JSON connection info file",
    default=None,
    description="""
        If the file does not exist, kernel connection information will be written
        to the file path provided.

        If the file exists, kernel connection info will be read from the file,
        allowing euporie to connect to existing kernels.
    """,
)



"""Define a configuration class for euporie.core."""

from __future__ import annotations

import argparse
import json
import logging
import os
import sys
from ast import literal_eval
from functools import cached_property, partial
from pathlib import Path
from types import SimpleNamespace
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Optional,
    TextIO,
    cast,
)

import fastjsonschema
from platformdirs import user_config_dir
from prompt_toolkit.filters.base import Condition
from prompt_toolkit.filters.utils import to_filter
from prompt_toolkit.utils import Event
from upath import UPath

from euporie.core import __app_name__, __copyright__
from euporie.core.commands import add_cmd, get_cmd

if TYPE_CHECKING:
    from collections.abc import Iterable, Mapping, Sequence
    from typing import IO, Any, Callable, ClassVar, Optional

    from prompt_toolkit.filters.base import FilterOrBool

    from euporie.core.widgets.menu import MenuItem


log = logging.getLogger(__name__)

_SCHEMA_TYPES: dict[type | Callable, str] = {
    bool: "boolean",
    str: "string",
    int: "integer",
    float: "number",
    UPath: "string",
}


class ArgumentParser(argparse.ArgumentParser):
    """An argument parser which lexes and formats the help message before printing it."""

    def __init__(self, *args: Any, config: Config, **kwargs: Any) -> None:
        """Initialize while saving a reference to the current config."""
        super().__init__(*args, **kwargs)
        self.config = config

    def _print_message(self, message: str, file: IO[str] | None = None) -> None:
        from prompt_toolkit.formatted_text.base import FormattedText
        from prompt_toolkit.lexers.pygments import _token_cache
        from prompt_toolkit.shortcuts.utils import print_formatted_text
        from prompt_toolkit.styles.pygments import style_from_pygments_cls

        from euporie.core.pygments import ArgparseLexer
        from euporie.core.style import get_style_by_name

        if message:
            file = cast("Optional[TextIO]", file)
            style = style_from_pygments_cls(get_style_by_name(self.config.syntax_theme))
            print_formatted_text(
                FormattedText(
                    [
                        (_token_cache[t], v)
                        for _, t, v in ArgparseLexer().get_tokens_unprocessed(
                            message.rstrip("\n")
                        )
                    ]
                ),
                file=file,
                style=style,
                include_default_pygments_style=False,
            )


class BooleanOptionalAction(argparse.Action):
    """Action for boolean flags.

    Included because `argparse.BooleanOptionalAction` is not present in `python<=3.9`.
    """

    def __init__(self, option_strings: list[str], *args: Any, **kwargs: Any) -> None:
        """Initiate the Action, as per `argparse.BooleanOptionalAction`."""
        _option_strings = list(option_strings)
        for option_string in option_strings:
            if option_string.startswith("--"):
                _option_strings.append(f"--no-{option_string[2:]}")
        kwargs["nargs"] = 0
        super().__init__(_option_strings, *args, **kwargs)

    def format_usage(self) -> str:
        """Format the action string.

        Returns:
            The formatted string.

        """
        return " | ".join(self.option_strings)

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: str | Sequence[Any] | None,
        option_string: str | None = None,
    ) -> None:
        """Set the value to True or False depending on the flag provided."""
        if option_string in self.option_strings:
            assert isinstance(option_string, str)
            setattr(namespace, self.dest, not option_string.startswith("--no-"))


TYPE_ACTIONS: dict[Callable[[Any], Any], type[argparse.Action]] = {
    bool: BooleanOptionalAction
}


class JSONEncoderPlus(json.JSONEncoder):
    """JSON encode class which encodes paths as strings."""

    def default(self, o: Any) -> bool | int | float | str | None:
        """Encode an object to JSON.

        Args:
            o: The object to encode

        Returns:
            The encoded object

        """
        if isinstance(o, (Path, UPath)):
            return str(o)
        return json.JSONEncoder.default(self, o)


_json_encoder = JSONEncoderPlus()


class Setting:
    """A single configuration item."""

    def __init__(
        self,
        name: str,
        group: str | Sequence[str],
        default: Any = None,
        help_: str = "",
        description: str = "",
        type_: Callable[[Any], Any] | None = None,
        title: str | None = None,
        choices: list[Any] | Callable[[], list[Any]] | None = None,
        action: argparse.Action | str | None = None,
        flags: list[str] | None = None,
        schema: dict[str, Any] | None = None,
        nargs: str | int | None = None,
        hidden: FilterOrBool = False,
        hooks: list[Callable[[Setting], None]] | None = None,
        cmd_filter: FilterOrBool = True,
        **kwargs: Any,
    ) -> None:
        """Create a new configuration item."""
        self.name = name
        self.groups = {group} if isinstance(group, str) else set(group)
        self.default = default
        self._value = default
        self.title = title or self.name.replace("_", " ")
        self.help = help_
        self.description = description
        self._choices = choices
        self.type = type_ or type(default)
        self.action = action or TYPE_ACTIONS.get(self.type)
        self.flags = flags or [f"--{name.replace('_', '-')}"]
        self._schema: dict[str, Any] = {
            "type": _SCHEMA_TYPES.get(self.type),
            **(schema or {}),
        }
        self.nargs = nargs
        self.hidden = to_filter(hidden)
        self.hooks = hooks
        self.cmd_filter = cmd_filter
        self.kwargs = kwargs

    @cached_property
    def choices(self) -> list[Any] | None:
        """Compute the setting's available options."""
        if callable(self._choices):
            return self._choices()
        else:
            return self._choices

    @property
    def schema(self) -> dict[str, Any]:
        """Return a json schema property for the config item."""
        schema = {
            "description": self.help,
            **({"default": self.default} if self.default is not None else {}),
            **self._schema,
        }
        if self.choices:
            if self.nargs == "*" or "items" in schema:
                schema["items"]["enum"] = self.choices
            else:
                schema["enum"] = self.choices
        return schema

    @property
    def menu(self) -> MenuItem:
        """Return a menu item for the setting."""
        from euporie.core.widgets.menu import MenuItem

        choices = (self.choices or self.schema.get("enum", [])) or []
        if choices:
            return MenuItem(
                self.title.capitalize(),
                children=[
                    cmd.menu
                    for cmd in sorted(
                        (
                            get_cmd(f"set-{self.name.replace('_', '-')}-{choice}")
                            for choice in choices
                        ),
                        key=lambda x: x.menu_title,
                    )
                ],
                description=self.help,
            )
        elif self.type in (bool, int):
            return get_cmd(f"toggle-{self.name.replace('_', '-')}").menu
        else:
            raise NotImplementedError

    @property
    def parser_args(self) -> tuple[list[str], dict[str, Any]]:
        """Return arguments for construction of an :class:`argparse.ArgumentParser`."""
        # If empty flags are passed, do not expose setting on command line
        if self.flags is not None and not self.flags:
            return [], {}
        args = self.flags or [self.name]

        kwargs: dict[str, Any] = {
            "action": self.action,
            "help": argparse.SUPPRESS if self.hidden() else self.help,
            # Do not set defaults for command line arguments, as default values
            # would override values set in the configuration file
            # "default": self.default,
        }

        if self.nargs:
            kwargs["nargs"] = self.nargs
        if self.type is not None and self.name != "version":
            kwargs["type"] = self.type
        if self.choices:
            kwargs["choices"] = self.choices
        if "version" in self.kwargs:
            kwargs["version"] = self.kwargs["version"]
        if "const" in self.kwargs:
            kwargs["const"] = self.kwargs["const"]

        return args, kwargs

    def __repr__(self) -> str:
        """Represent a :py:class`Setting` instance as a string."""
        return f"<Setting {self.name}: {self.type}>"


class DefaultNamespace(SimpleNamespace):
    """A namespace that creates default values for undefined attributes using a factory function.

    This class extends SimpleNamespace to provide automatic creation of default values when
    accessing undefined attributes, similar to collections.defaultdict but for object attributes.

    Attributes:
        _factory: A callable that generates default values for undefined attributes.
            If None, AttributeError will be raised for undefined attributes.

    Examples:
        >>> # Create namespace with list factory
        >>> ns = DefaultNamespace(default_factory=list)
        >>> ns.numbers.append(1)  # Creates new list automatically
        >>> ns.numbers
        [1]

        >>> # Create with initial values
        >>> ns = DefaultNamespace(default_factory=int, x=1, y=2)
        >>> ns.x
        1
        >>> ns.z  # Creates new int (0) automatically
        0
    """

    def __init__(
        self,
        default_factory: Callable[[str], Any] | None = None,
        mapping_or_iterable: Mapping | Iterable[tuple[str, Any]] | None = None,
        /,
        **kwargs: Any,
    ) -> None:
        """Initialize the DefaultNamespace.

        Args:
            default_factory: A callable that takes an attribute name and returns a default value
                when accessing undefined attributes. If None, AttributeError is raised for
                undefined attributes.
            mapping_or_iterable: An optional mapping or iterable of (key, value) pairs to
                initialize the namespace with.
            **kwargs: Additional keyword arguments to initialize the namespace with.
        """
        if mapping_or_iterable is None:
            mapping_or_iterable = {}
        super().__init__(**dict(mapping_or_iterable), **kwargs)
        self._factory = default_factory

    def __getattribute__(self, name: str) -> Any:
        """Get an attribute value, creating it if undefined using the default_factory.

        This method intercepts attribute access to provide default value creation
        for undefined attributes using the default_factory.

        Args:
            name: The name of the attribute to access.

        Returns:
            The attribute value, either existing or newly created.

        Raises:
            AttributeError: If the attribute doesn't exist and no default_factory is set.
        """
        try:
            return super().__getattribute__(name)
        except AttributeError:
            factory = super().__getattribute__("_factory")
            if factory is None:
                raise
            value = factory(name)
            setattr(self, name, value)
            return value


class Config:
    """A configuration store."""

    _conf_file_name = "config.json"
    _settings: ClassVar[dict[str, Setting]] = {}

    def __init__(self, _help: str = "", **kwargs: Any) -> None:
        """Create a new configuration object instance."""
        self._help = _help

        # Create stores
        self.events = DefaultNamespace(
            lambda name: Event(self._settings[name], self._save)
        )
        self.filters = DefaultNamespace(
            lambda name: Condition(lambda: bool(self._values[name]))
        )
        self.defaults = DefaultNamespace(lambda x: self._settings[x].default)
        self.choices = DefaultNamespace(lambda x: self._settings[x].choices)
        self.menus = DefaultNamespace(lambda x: self._settings[x].menu)

        user_conf_dir = Path(user_config_dir(__app_name__, appauthor=None))
        user_conf_dir.mkdir(exist_ok=True, parents=True)
        self._config_file_path = (user_conf_dir / self._conf_file_name).with_suffix(
            ".json"
        )
        self._valid_user = True
        self._schema_validate = fastjsonschema.compile(self._schema, use_default=False)

        # Register commands for each setting
        for setting in self._settings.values():
            self._register_commands(setting)

        self._values = {
            # Setting defaults
            **{k: v.default for k, v in self._settings.items()},
            # Key-word arguments
            **kwargs,
        }

    def load(self) -> None:
        """Load the configuration options from non-local sources."""
        from euporie.core.log import BufferedLogs, setup_logs

        # Buffer logs and replay them after settings are configured
        with BufferedLogs(logger=log):
            try:
                # Validate configured values
                self._values.update(self._validate(self._load_user(), "config file"))
                self._values.update(
                    self._validate(self._load_user(prefix=self.app), "app config file")
                )
                self._values.update(
                    self._validate(self._load_env(), "environment variable")
                )
                self._values.update(
                    self._validate(
                        self._load_env(prefix=self.app), "app environment variable"
                    )
                )
                self._values.update(
                    self._validate(self._load_args(), "command line parameter")
                )
            finally:
                # Set-up logs even if configuration validation fails
                setup_logs(self)

    def _validate(self, data: dict[str, Any], group: str) -> dict[str, Any]:
        """Validate settings values."""
        validated = {}
        for name, value in data.items():
            if name in self._settings:
                # Convert to json and back to attain json types
                json_data = json.loads(_json_encoder.encode({name: value}))
                try:
                    self._schema_validate(json_data)
                except fastjsonschema.JsonSchemaValueException as error:
                    # Warn about badly configured settings
                    log.warning(
                        "Error in %s setting: `%s = %r`\n%s",
                        group,
                        name,
                        value,
                        error.message.replace("data.", ""),
                    )
                else:
                    # Store validated setting values
                    validated[name] = value
            else:
                # Warn about unknown configuration options
                if not isinstance(value, dict):
                    log.warning(
                        "Configuration option '%s' not recognised in %s", name, group
                    )
        return validated

    def _save(self, setting: Setting) -> None:
        """Save settings to user's configuration file."""
        json_data = self._load_config_file()
        json_data.setdefault(self.app, {})[setting.name] = getattr(self, setting.name)
        if self._valid_user:
            log.debug("Saving setting `%s`", setting)
            with self._config_file_path.open("w") as f:
                json.dump(json_data, f, indent=2)

    @property
    def _schema(self) -> dict[str, Any]:
        """Return a JSON schema for the config."""
        return {
            "title": "Euporie Configuration",
            "description": "A configuration for euporie",
            "type": "object",
            "properties": {name: item.schema for name, item in self._settings.items()},
        }

    @property
    def settings(self) -> dict[str, Setting]:
        """Return the currently active settings."""
        return {
            name: setting
            for name, setting in self._settings.items()
            if any(group in sys.modules or group == "*" for group in setting.groups)
        }

    def _load_parser(self) -> argparse.ArgumentParser:
        """Construct an :py:class:`ArgumentParser`."""
        parser = ArgumentParser(
            description=self._help,
            epilog=__copyright__,
            allow_abbrev=True,
            formatter_class=argparse.MetavarTypeHelpFormatter,
            config=self,
            argument_default=argparse.SUPPRESS,
            # exit_on_error=False,
        )
        # Add options to the relevant subparsers
        for setting in self.settings.values():
            # if self._values[setting.name] != setting.default:
            args, kwargs = setting.parser_args
            # Make already specified positional arguments optional
            if (
                not any(x.startswith("-") for x in setting.flags)
                and getattr(self, setting.name) != setting.default
            ):
                kwargs["nargs"] = "?"
            parser.add_argument(*args, **kwargs)
        return parser

    def _load_args(self) -> dict[str, Any]:
        """Attempt to load configuration settings from commandline flags."""
        # Parse known arguments
        namespace, remainder = self._load_parser().parse_known_intermixed_args()
        # Validate arguments
        return vars(namespace)

    def _load_env(self, prefix: str = "") -> dict[str, Any]:
        """Attempt to load configuration settings from environment variables."""
        result = {}
        for name, setting in self._settings.items():
            if prefix:
                env = f"{__app_name__}_{prefix}_{setting.name}"
            else:
                env = f"{__app_name__}_{setting.name}"
            env = env.upper()
            if env in os.environ:
                value = os.environ.get(env)
                parsed_value: Any = value
                # Attempt to parse the value as a literal
                if value:
                    try:
                        parsed_value = literal_eval(value)
                    except (
                        ValueError,
                        TypeError,
                        SyntaxError,
                        MemoryError,
                        RecursionError,
                    ):
                        pass
                # Attempt to cast the value to the desired type
                try:
                    parsed_value = setting.type(value)
                except (ValueError, TypeError):
                    pass
                result[name] = parsed_value
        return result

    def _load_config_file(self) -> dict[str, Any]:
        """Attempt to load JSON configuration file."""
        results = {}
        assert isinstance(self._config_file_path, Path)
        if self._valid_user and self._config_file_path.exists():
            with self._config_file_path.open() as f:
                try:
                    json_data = json.load(f)
                except json.decoder.JSONDecodeError:
                    log.error(
                        "Could not parse the configuration file: %s\nIs it valid json?",
                        self._config_file_path,
                    )
                    self._valid_user = False
                else:
                    results.update(json_data)
        return results

    def _load_user(self, prefix: str | None = None) -> dict[str, Any]:
        """Attempt to load JSON configuration file."""
        results = {}
        # Load config file
        json_data = self._load_config_file()
        # Load section for the current app
        if prefix is not None:
            json_data = json_data.get(prefix, {})
        for name, value in json_data.items():
            if (setting := self._settings.get(name)) is not None:
                # Attempt to cast the value to the desired type
                try:
                    value = setting.type(value)
                except (ValueError, TypeError):
                    pass
            results[name] = value
        return results

    def _register_commands(self, setting: Setting) -> None:
        """Register commands to set this setting."""
        cmd_name = setting.name.replace("_", "-")
        schema = setting.schema

        if schema.get("type") == "array":
            for choice in setting.choices or schema.get("items", {}).get("enum") or []:
                add_cmd(
                    name=f"toggle-{cmd_name}-{choice}",
                    hidden=setting.hidden,
                    toggled=Condition(
                        partial(lambda x: x in self._values[setting.name], choice)
                    ),
                    title=f"Add {choice} to {setting.title} setting",
                    menu_title=str(choice).replace("_", " ").capitalize(),
                    description=f'Add or remove "{choice}" to or from the list of "{setting.name}"',
                    filter=setting.cmd_filter,
                )(
                    partial(
                        lambda choice: (
                            self._values[setting.name].remove
                            if choice in getattr(self, setting.name)
                            else self._values[setting.name].append
                        )(choice),
                        choice,
                    )
                )

        elif setting.type is bool:
            add_cmd(
                name=f"toggle-{cmd_name}",
                toggled=Condition(lambda: bool(self._values[setting.name])),
                hidden=setting.hidden,
                title=f"Toggle {setting.title}",
                menu_title=setting.kwargs.get("menu_title", setting.title.capitalize()),
                description=setting.help,
                filter=setting.cmd_filter,
            )(partial(self.toggle, setting.name))

        elif setting.type is int or setting.choices is not None:
            add_cmd(
                name=f"switch-{cmd_name}",
                hidden=setting.hidden,
                title=f"Switch {setting.title}",
                menu_title=setting.kwargs.get("menu_title"),
                description=f'Switch the value of the "{setting.name}" configuration option.',
                filter=setting.cmd_filter,
            )(partial(self.toggle, setting.name))

        for choice in setting.choices or schema.get("enum", []) or []:
            add_cmd(
                name=f"set-{cmd_name}-{choice}",
                hidden=setting.hidden,
                toggled=Condition(
                    partial(lambda x: self._values[setting.name] == x, choice)
                ),
                title=f"Set {setting.title} to {choice}",
                menu_title=str(choice).replace("_", " ").capitalize(),
                description=f'Set the value of the "{setting.name}" '
                f'configuration option to "{choice}"',
                filter=setting.cmd_filter,
            )(partial(setattr, self, setting.name, choice))

    def toggle(self, name: str) -> None:
        """Toggle the setting's value."""
        setting = self._settings[name]
        value = self._values[name]

        if setting.type is bool:
            new = not value
        elif (
            setting.type is int
            and "minimum" in (schema := setting.schema)
            and "maximum" in schema
        ):
            new = schema["minimum"] + (value - schema["minimum"] + 1) % (
                schema["maximum"] + 1
            )
        elif setting.choices is not None:
            new = setting.choices[
                (setting.choices.index(value) + 1) % len(setting.choices)
            ]
        else:
            raise NotImplementedError
        setattr(self, name, new)

    def __getattribute__(self, name: str) -> Any:
        """Enable access of config elements via dotted attributes."""
        try:
            return super().__getattribute__(name)
        except AttributeError as exc:
            if name in self._values:
                return self._values[name]
            raise exc

    def __setattr__(self, name: str, value: Any) -> None:
        """Set a configuration attribute."""
        if name in self._settings:
            self._values[name] = value
            getattr(self.events, name)()
        else:
            super().__setattr__(name, value)

    @classmethod
    def add_setting(cls, name: str, *args: Any, **kwargs: Any) -> None:
        """Register a new config item."""
        setting = Setting(name, *args, **kwargs)
        Config._settings[name] = setting


add_setting = Config.add_setting



"""Contain functions which convert data to base64 format."""

from __future__ import annotations

import base64
from typing import TYPE_CHECKING

from euporie.core.convert.registry import register

if TYPE_CHECKING:
    from euporie.core.convert.datum import Datum


@register(from_="png", to="base64-png", filter_=True)
@register(from_="jpeg", to="base64-jpeg", filter_=True)
@register(from_="pdf", to="base64-pdf", filter_=True)
@register(from_="svg", to="base64-svg", filter_=True)
async def bytes_to_base64_py(
    datum: Datum,
    cols: int | None = None,
    rows: int | None = None,
    fg: str | None = None,
    bg: str | None = None,
    extend: bool = True,
) -> str:
    """Convert bytes to base64 encoded data."""
    data = datum.data
    if isinstance(data, str):
        data = data.encode()
    return base64.b64encode(data).decode().replace("\n", "").strip()



"""Defines a container which displays all children at full height vertially stacked."""

from __future__ import annotations

import logging
from functools import lru_cache
from typing import TYPE_CHECKING

from prompt_toolkit.filters import Condition
from prompt_toolkit.layout.containers import (
    Container,
    to_container,
)
from prompt_toolkit.mouse_events import MouseEventType

from euporie.core.app.current import get_app

if TYPE_CHECKING:
    from typing import Callable

    from prompt_toolkit.key_binding.key_bindings import NotImplementedOrNone
    from prompt_toolkit.layout.containers import AnyContainer
    from prompt_toolkit.layout.dimension import Dimension
    from prompt_toolkit.layout.mouse_handlers import MouseHandlers
    from prompt_toolkit.layout.screen import Screen, WritePosition
    from prompt_toolkit.mouse_events import MouseEvent

    MouseHandler = Callable[[MouseEvent], object]

log = logging.getLogger(__name__)


class DisableMouseOnScroll(Container):
    """A container which disables mouse support on unhandled scroll up events.

    This enables the terminal scroll-back buffer to be scrolled if there is nothing in
    the application which is scrollable.
    """

    def __init__(
        self,
        content: AnyContainer,
    ) -> None:
        """Initiate the container."""
        self.content = to_container(content)

        # Wrap app mouse support filter
        app = get_app()
        self.mouse_support = app.mouse_support()
        filter_ = Condition(lambda: self.mouse_support)
        app.mouse_support &= filter_
        app.renderer.mouse_support &= filter_

    def reset(self) -> None:
        """Reset the state of this container."""
        self.content.reset()

    def preferred_width(self, max_available_width: int) -> Dimension:
        """Return the desired width for this container."""
        return self.content.preferred_width(max_available_width)

    def preferred_height(self, width: int, max_available_height: int) -> Dimension:
        """Return the desired height for this container."""
        return self.content.preferred_height(width, max_available_height)

    def write_to_screen(
        self,
        screen: Screen,
        mouse_handlers: MouseHandlers,
        write_position: WritePosition,
        parent_style: str,
        erase_bg: bool,
        z_index: int | None,
    ) -> None:
        """Render the container to a `Screen` instance.

        Wrap mouse handelrs, hooking unhandled scroll up events so the terminal can be
        scrolled if the scroll event is not handled.

        Args:
            screen: The :class:`~prompt_toolkit.layout.screen.Screen` class to which
                the output has to be written.
            mouse_handlers: :class:`prompt_toolkit.layout.mouse_handlers.MouseHandlers`.
            write_position: A :class:`prompt_toolkit.layout.screen.WritePosition` object
                defining where this container should be drawn.
            erase_bg: If true, the background will be erased prior to drawing.
            parent_style: Style string to pass to the :class:`.Window` object. This will
                be applied to all content of the windows. :class:`.VSplit` and
                :class:`prompt_toolkit.layout.containers.HSplit` can use it to pass
                their style down to the windows that they contain.
            z_index: Used for propagating z_index from parent to child.

        """
        self.mouse_support = True
        self.content.write_to_screen(
            screen,
            mouse_handlers,
            write_position,
            parent_style,
            erase_bg,
            z_index,
        )

        @lru_cache
        def _wrap_mouse_handler(handler: Callable) -> MouseHandler:
            def wrapped_mouse_handler(mouse_event: MouseEvent) -> NotImplementedOrNone:
                response = handler(mouse_event)
                # Disable mouse support if scrolling is not implemented
                if (
                    mouse_event.event_type == MouseEventType.SCROLL_UP
                    and response is NotImplemented
                ):
                    self.mouse_support = False
                    return None

                    # async def _reactivate_mouse_support() -> None:
                    #     import asyncio
                    #     await asyncio.sleep(1)
                    #     get_app().need_mouse_support = True
                    #     get_app().invalidate()
                    # get_app().create_background_task(_reactivate_mouse_support())

                return response

            return wrapped_mouse_handler

        # Wrap mouse handlers
        mhs = mouse_handlers.mouse_handlers
        for y in range(
            write_position.ypos, write_position.ypos + write_position.height
        ):
            row = mhs[y]
            for x in range(
                write_position.xpos, write_position.xpos + write_position.width
            ):
                row[x] = _wrap_mouse_handler(row[x])

    def get_children(self) -> list[Container]:
        """Return a list of all child containers."""
        return [self.content]



"""Defines a web-view control."""

from __future__ import annotations

import asyncio
import logging
from threading import Thread
from typing import TYPE_CHECKING, cast

from prompt_toolkit.cache import FastDictCache
from prompt_toolkit.data_structures import Point
from prompt_toolkit.filters import Condition
from prompt_toolkit.formatted_text.utils import split_lines
from prompt_toolkit.layout.containers import Window
from prompt_toolkit.layout.controls import UIContent, UIControl
from prompt_toolkit.mouse_events import MouseButton, MouseEvent, MouseEventType
from prompt_toolkit.utils import Event
from upath import UPath

from euporie.core.app.current import get_app
from euporie.core.commands import add_cmd
from euporie.core.convert.datum import Datum, get_loop
from euporie.core.convert.mime import get_format
from euporie.core.ft.html import HTML, Node
from euporie.core.ft.utils import fragment_list_width, paste
from euporie.core.graphics import GraphicProcessor
from euporie.core.key_binding.registry import (
    load_registered_bindings,
    register_bindings,
)
from euporie.core.path import parse_path

if TYPE_CHECKING:
    from collections.abc import Iterable
    from concurrent.futures._base import Future
    from pathlib import Path
    from typing import Any, Callable

    from prompt_toolkit.formatted_text.base import AnyFormattedText, StyleAndTextTuples
    from prompt_toolkit.key_binding.key_bindings import (
        KeyBindingsBase,
        NotImplementedOrNone,
    )
    from prompt_toolkit.layout.controls import GetLinePrefixCallable
    from prompt_toolkit.layout.mouse_handlers import MouseHandler


log = logging.getLogger(__name__)


@Condition
def webview_has_focus() -> bool:
    """Determine if there is a currently focused webview."""
    return isinstance(get_app().layout.current_control, WebViewControl)


class WebViewControl(UIControl):
    """Web view displays.

    A control which displays rendered HTML content.
    """

    _window: Window

    def __init__(
        self,
        url: str | Path = "about:blank",
        link_handler: Callable | None = None,
    ) -> None:
        """Create a new web-view control instance."""
        self._cursor_position = Point(0, 0)
        self.loading = False
        self.resizing = False
        self.rendering = False
        self.stale = False
        self.lines: list[StyleAndTextTuples] = []
        self.graphic_processor = GraphicProcessor(control=self)
        self.width = 0
        self.height = 0
        self.url: Path = UPath(url)
        self.status: list[AnyFormattedText] = []
        self.link_handler = link_handler or self.load_url

        self.render_task: Future[None] | None = None
        self.rendered = Event(self)
        self.on_cursor_position_changed = Event(self)

        self.prev_stack: list[Path] = []
        self.next_stack: list[Path] = []

        # self.cursor_processor = CursorProcessor(
        #     lambda: self.cursor_position, style="fg:red"
        # )

        self.loop = asyncio.new_event_loop()
        self.render_thread = Thread(target=self.loop.run_forever, daemon=True)

        self.key_bindings = load_registered_bindings(
            "euporie.web.widgets.webview.WebViewControl",
            config=get_app().config,
        )

        self._dom_cache: FastDictCache[tuple[Path], HTML] = FastDictCache(
            get_value=self.get_dom, size=100
        )
        self._content_cache: FastDictCache = FastDictCache(self.get_content, size=1_000)

        self.load_url(url)

    @property
    def window(self) -> Window:
        """Get the control's window."""
        try:
            return self._window
        except AttributeError:
            for window in get_app().layout.find_all_windows():
                if window.content == self:
                    self._window = window
                    return window
        return Window()

    @property
    def cursor_position(self) -> Point:
        """Get the cursor position."""
        return self._cursor_position

    @cursor_position.setter
    def cursor_position(self, value: Point) -> None:
        """Set the cursor position."""
        changed = self._cursor_position != value
        self._cursor_position = value
        if changed:
            self.on_cursor_position_changed.fire()

    def get_dom(self, url: Path) -> HTML:
        """Load a HTML page as renderable formatted text."""
        markup = str(
            Datum(
                data=url.read_text(),
                format=(format_ := get_format(url, default="html")),
                path=url,
            ).convert(to="html")
        )
        return HTML(
            markup=markup,
            base=url,
            mouse_handler=self._node_mouse_handler,
            paste_fixed=False,
            _initial_format=format_,
            defer_assets=False,
            on_change=lambda dom: self.invalidate(),
        )

    def invalidate(self) -> None:
        """Trigger a redraw of the webview."""
        self.stale = True
        self.rendered.fire()

    @property
    def dom(self) -> HTML:
        """Return the dom for the current URL."""
        return self._dom_cache[self.url,]

    @property
    def title(self) -> str:
        """Return the title of the current HTML page."""
        if self.loading:
            return self.url.name
        if dom := self.dom:
            return dom.title
        return ""

    def load_url(self, url: str | Path, **kwargs: Any) -> None:
        """Load a new URL."""
        save_to_history = kwargs.get("save_to_history", True)
        # Trigger "loading" view
        self.loading = True
        # Update navigation history
        if self.url and save_to_history:
            self.prev_stack.append(self.url)
            self.next_stack.clear()
        # Update url
        self.url = UPath(url)
        # Scroll to the top
        # self.window.vertical_scroll = 0  # TODO - save scroll in history
        self.cursor_position = Point(0, 0)
        # Signal that the webview has updated
        self.rendered.fire()

    def nav_prev(self) -> None:
        """Navigate forwards through the browser history."""
        if self.url and self.prev_stack:
            self.next_stack.append(self.url)
            self.load_url(self.prev_stack.pop(), save_to_history=False)

    def nav_next(self) -> None:
        """Navigate backwards through the browser history."""
        if self.url and self.next_stack:
            self.prev_stack.append(self.url)
            self.load_url(self.next_stack.pop(), save_to_history=False)

    def render(self, force: bool = False) -> None:
        """Render the HTML DOM in a thread."""
        dom = self.dom

        async def _render() -> None:
            assert self.url is not None
            # Potentially redirect url
            self.url = parse_path(self.url)
            self.lines = list(split_lines(await dom._render(self.width, self.height)))
            # Reset all possible reasons for rendering
            self.loading = self.resizing = self.rendering = self.stale = False
            # Let the app know we're re-rerenderd and the output needs redrawing
            self.rendered.fire()

        if not self.rendering or force:
            self.rendering = True

            if self.render_task:
                self.render_task.cancel()

            self.render_task = asyncio.run_coroutine_threadsafe(_render(), get_loop())

    def reset(self) -> None:
        """Reset the state of the control."""

    def preferred_width(self, max_available_width: int) -> int | None:
        """Calculate and return the preferred width of the control."""
        return None

    def preferred_height(
        self,
        width: int,
        max_available_height: int,
        wrap_lines: bool,
        get_line_prefix: GetLinePrefixCallable | None,
    ) -> int | None:
        """Calculate and return the preferred height of the control."""
        return None

    def is_focusable(self) -> bool:
        """Tell whether this user control is focusable."""
        return True

    def get_content(
        self,
        url: Path,
        loading: bool,
        resizing: bool,
        width: int,
        height: int,
        cursor_position: Point,
        assets_loaded: bool,
    ) -> UIContent:
        """Create a cacheable UIContent."""
        dom = self._dom_cache[url,]
        if self.loading:
            lines = [
                cast("StyleAndTextTuples", []),
                cast(
                    "StyleAndTextTuples",
                    [("", " " * ((width - 8) // 2)), ("class:loading", "Loading…")],
                ),
            ]
        else:
            lines = self.lines[:]

        def get_line(i: int) -> StyleAndTextTuples:
            try:
                line = lines[i]
            except IndexError:
                line = []

            # Overlay fixed lines onto this line
            if not loading and dom.fixed:
                visible_line = max(0, i - self.window.vertical_scroll)
                fixed_lines = list(split_lines(dom.fixed_mask))
                if visible_line < len(fixed_lines):
                    # Paste the fixed line over the current line
                    fixed_line = fixed_lines[visible_line]
                    line = paste(fixed_line, line, 0, 0, transparent=True)

            return line

        return UIContent(
            get_line=get_line,
            line_count=len(lines),
            cursor_position=self.cursor_position,
            show_cursor=False,
        )

    def create_content(self, width: int, height: int) -> UIContent:
        """Generate the content for this user control.

        Returns:
            A :class:`.UIContent` instance.
        """
        # Trigger a re-render in the future if things have changed
        if self.stale or self.loading:
            self.render()
        if width != self.width:  # or height != self.height:
            # self.resizing = True
            self.width = width
            self.height = height
            self.render()

        content = self._content_cache[
            self.url,
            self.loading,
            self.resizing,
            width,
            height,
            self.cursor_position,
            self.dom.render_count,
        ]

        # Check for graphics in content
        self.graphic_processor.load(content)

        return content

    def _node_mouse_handler(
        self, node: Node, mouse_event: MouseEvent
    ) -> NotImplementedOrNone:
        """Handle click events."""
        url = node.attrs.get("_link_path")
        title = node.attrs.get("title")
        alt = node.attrs.get("alt")
        if url:
            # TODO - Check for #anchor links and scroll accordingly
            if (
                mouse_event.button == MouseButton.LEFT
                and mouse_event.event_type == MouseEventType.MOUSE_UP
            ):
                self.link_handler(url, save_to_history=True, new_tab=False)
                return None
            elif (
                mouse_event.button == MouseButton.MIDDLE
                and mouse_event.event_type == MouseEventType.MOUSE_UP
            ):
                self.link_handler(url, save_to_history=False, new_tab=True)
                return None
        if (url or title) and mouse_event.event_type == MouseEventType.MOUSE_MOVE:
            self.status.clear()
            if title:
                self.status.append(str(title))
            if title:
                self.status.append(str(alt))
            if url:
                self.status.append(str(url))
            return None
        return NotImplemented

    def mouse_handler(self, mouse_event: MouseEvent) -> NotImplementedOrNone:
        """Handle mouse events.

        When `NotImplemented` is returned, it means that the given event is not
        handled by the `UIControl` itself. The `Window` or key bindings can
        decide to handle this event as scrolling or changing focus.

        Args:
            mouse_event: `MouseEvent` instance.

        Returns:
            NotImplemented if the UI does not need to be updates, None if it does
        """
        # Focus on mouse down
        if mouse_event.event_type == MouseEventType.MOUSE_DOWN:
            if not (layout := get_app().layout).has_focus(self):
                layout.focus(self)
                return None
            return NotImplemented
        # if mouse_event.event_type == MouseEventType.MOUSE_MOVE:
        # self.cursor_position = mouse_event.position

        # if mouse_event.event_type == MouseEventType.MOUSE_UP:
        handler: MouseHandler | None = None
        try:
            content = self._content_cache[
                self.url,
                self.loading,
                self.resizing,
                self.width,
                self.height,
                self.cursor_position,
                self.dom.render_count,
            ]
            line = content.get_line(mouse_event.position.y)
        except IndexError:
            return NotImplemented
        else:
            # Find position in the fragment list.
            xpos = mouse_event.position.x
            # Find mouse handler for this character.
            count = 0
            for item in line:
                count += len(item[1])
                if count > xpos:
                    if len(item) >= 3:
                        handler = item[2]
                        if callable(handler):
                            return handler(mouse_event)
                    else:
                        break

        if callable(handler):
            return handler

        if self.status:
            self.status.clear()
            return None

        return NotImplemented

    @property
    def content_width(self) -> int:
        """Return the width of the content."""
        return max(fragment_list_width(line) for line in self.lines)

    def move_cursor_down(self) -> None:
        """Move the cursor down one line."""
        x, y = self.cursor_position
        self.cursor_position = Point(x=x, y=y + 1)

    def move_cursor_up(self) -> None:
        """Move the cursor up one line."""
        x, y = self.cursor_position
        self.cursor_position = Point(x=x, y=max(0, y - 1))

    def move_cursor_left(self) -> None:
        """Move the cursor down one line."""
        x, y = self.cursor_position
        self.cursor_position = Point(x=max(0, x - 1), y=y)

    def move_cursor_right(self) -> None:
        """Move the cursor up one line."""
        x, y = self.cursor_position
        self.cursor_position = Point(x=x + 1, y=y)

    def get_key_bindings(self) -> KeyBindingsBase | None:
        """Return key bindings that are specific for this user control.

        Returns:
            A :class:`.KeyBindings` object if some key bindings are specified, or
                `None` otherwise.
        """
        return self.key_bindings

    def get_invalidate_events(self) -> Iterable[Event[object]]:
        """Return a list of `Event` objects, which can be a generator.

        the application collects all these events, in order to bind redraw
        handlers to these events.
        """
        yield self.rendered
        yield self.on_cursor_position_changed
        if dom := self.dom:
            yield dom.on_update

    # ################################### Commands ####################################

    @staticmethod
    @add_cmd(filter=webview_has_focus)
    def _webview_nav_prev() -> None:
        """Navigate backwards in the browser history."""
        from euporie.web.widgets.webview import WebViewControl

        current_control = get_app().layout.current_control
        if isinstance(current_control, WebViewControl):
            current_control.nav_prev()

    @staticmethod
    @add_cmd(filter=webview_has_focus)
    def _webview_nav_next() -> None:
        """Navigate forwards in the browser history."""
        from euporie.web.widgets.webview import WebViewControl

        current_control = get_app().layout.current_control
        if isinstance(current_control, WebViewControl):
            current_control.nav_next()

    @staticmethod
    @add_cmd(filter=webview_has_focus)
    def _scroll_webview_left() -> None:
        """Scroll the display up one line."""
        from euporie.core.widgets.display import DisplayWindow

        window = get_app().layout.current_window
        assert isinstance(window, DisplayWindow)
        window._scroll_left()

    @staticmethod
    @add_cmd(filter=webview_has_focus)
    def _scroll_webview_right() -> None:
        """Scroll the display down one line."""
        from euporie.core.widgets.display import DisplayWindow

        window = get_app().layout.current_window
        assert isinstance(window, DisplayWindow)
        window._scroll_right()

    @staticmethod
    @add_cmd(filter=webview_has_focus)
    def _scroll_webview_up() -> None:
        """Scroll the display up one line."""
        get_app().layout.current_window._scroll_up()

    @staticmethod
    @add_cmd(filter=webview_has_focus)
    def _scroll_webview_down() -> None:
        """Scroll the display down one line."""
        get_app().layout.current_window._scroll_down()

    @staticmethod
    @add_cmd(filter=webview_has_focus)
    def _page_up_webview() -> None:
        """Scroll the display up one page."""
        window = get_app().layout.current_window
        if window.render_info is not None:
            for _ in range(window.render_info.window_height):
                window._scroll_up()

    @staticmethod
    @add_cmd(filter=webview_has_focus)
    def _page_down_webview() -> None:
        """Scroll the display down one page."""
        window = get_app().layout.current_window
        if window.render_info is not None:
            for _ in range(window.render_info.window_height):
                window._scroll_down()

    @staticmethod
    @add_cmd(filter=webview_has_focus)
    def _go_to_start_of_webview() -> None:
        """Scroll the display to the top."""
        from euporie.web.widgets.webview import WebViewControl

        current_control = get_app().layout.current_control
        if isinstance(current_control, WebViewControl):
            current_control.cursor_position = Point(0, 0)

    @staticmethod
    @add_cmd(filter=webview_has_focus)
    def _go_to_end_of_webview() -> None:
        """Scroll the display down one page."""
        from euporie.web.widgets.webview import WebViewControl

        layout = get_app().layout
        current_control = layout.current_control
        window = layout.current_window
        if (
            isinstance(current_control, WebViewControl)
            and window.render_info is not None
        ):
            current_control.cursor_position = Point(
                0, window.render_info.ui_content.line_count - 1
            )

    # ################################# Key Bindings ##################################

    register_bindings(
        {
            "euporie.web.widgets.webview.WebViewControl": {
                "scroll-webview-left": "left",
                "scroll-webview-right": "right",
                "scroll-webview-up": ["up", "k"],
                "scroll-webview-down": ["down", "j"],
                "page-up-webview": "pageup",
                "page-down-webview": "pagedown",
                "go-to-start-of-webview": "home",
                "go-to-end-of-webview": "end",
                "webview-nav-prev": ("A-left"),
                "webview-nav-next": ("A-right"),
            }
        }
    )


if __name__ == "__main__":
    import sys

    from prompt_toolkit.application.application import Application
    from prompt_toolkit.key_binding.key_bindings import KeyBindings
    from prompt_toolkit.layout.layout import Layout
    from prompt_toolkit.output.color_depth import ColorDepth
    from prompt_toolkit.styles.style import Style

    from euporie.core.style import HTML_STYLE
    from euporie.core.widgets.display import DisplayWindow
    from euporie.web.widgets.webview import WebViewControl

    kb = KeyBindings()
    kb.add("q")(lambda event: event.app.exit())
    layout = Layout(container=DisplayWindow(WebViewControl(UPath(sys.argv[-1]))))
    app: Application = Application(
        layout=layout,
        key_bindings=kb,
        full_screen=True,
        style=Style(HTML_STYLE),
        mouse_support=True,
        color_depth=ColorDepth.DEPTH_24_BIT,
    )
    app.run()



"""Contain the base class for a notebook tabs."""

from __future__ import annotations

import logging
from abc import ABCMeta, abstractmethod, abstractproperty
from base64 import standard_b64decode
from functools import partial
from typing import TYPE_CHECKING

import nbformat
from prompt_toolkit.filters import Never

from euporie.core.comm.registry import open_comm
from euporie.core.io import edit_in_editor
from euporie.core.kernel.base import MsgCallbacks
from euporie.core.tabs.kernel import KernelTab
from euporie.core.widgets.cell import Cell, get_cell_id

try:
    from jupytext import read as read_nb
    from jupytext import write as write_nb
except ModuleNotFoundError:
    from nbformat import read as read_nb
    from nbformat import write as write_nb

if TYPE_CHECKING:
    from collections.abc import Sequence
    from pathlib import Path
    from typing import Any, Callable

    from prompt_toolkit.filters import Filter
    from prompt_toolkit.layout.containers import AnyContainer
    from prompt_toolkit.layout.controls import BufferControl

    from euporie.core.app.app import BaseApp
    from euporie.core.comm.base import Comm
    from euporie.core.kernel.base import BaseKernel
    from euporie.core.lsp import LspClient
    from euporie.core.widgets.inputs import KernelInput

log = logging.getLogger(__name__)


class BaseNotebook(KernelTab, metaclass=ABCMeta):
    """The main notebook container class."""

    allow_stdin = False
    edit_mode = False

    def __init__(
        self,
        app: BaseApp,
        path: Path | None = None,
        kernel: BaseKernel | None = None,
        comms: dict[str, Comm] | None = None,
        use_kernel_history: bool = False,
        json: dict[str, Any] | None = None,
    ) -> None:
        """Instantiate a Notebook container, using a notebook at a given path.

        Args:
            path: The file path of the notebook
            app: The euporie application the notebook tab belongs to
            kernel: An existing kernel instance to use
            comms: Existing kernel comm object to use
            use_kernel_history: If True, load history from the kernel
            json: JSON contents of notebook, used if path not given

        """
        self.default_callbacks = MsgCallbacks(
            {
                "get_input": lambda prompt, password: self.cell.get_input(
                    prompt, password
                ),
                "set_execution_count": lambda n: self.cell.set_execution_count(n),
                "add_output": self.new_output_default,
                "clear_output": lambda wait: self.cell.clear_output(wait),
                "set_metadata": lambda path, data: self.cell.set_metadata(path, data),
                "set_status": self.set_status,
                "set_kernel_info": self.set_kernel_info,
                "dead": self.kernel_died,
                "edit_magic": edit_in_editor,
            }
        )
        self.json = json or {}
        self._rendered_cells: dict[str, Cell] = {}
        self.multiple_cells_selected: Filter = Never()
        self.loaded = False

        super().__init__(
            app, path, kernel=kernel, comms=comms, use_kernel_history=use_kernel_history
        )

        # Load notebook file
        self.container = self.load_container()

    # Tab stuff

    def reset(self) -> None:
        """Reload the notebook file from the disk and re-render."""
        # Restore selection after reset
        if self.path is not None:
            self._rendered_cells = {}
            self.load()
        self.refresh()

    # KernelTab stuff

    def post_init_kernel(self) -> None:
        """Load the notebook container after the kernel has been loaded."""
        super().post_init_kernel()

        # Load widgets
        self.load_widgets_from_metadata()

    @property
    def metadata(self) -> dict[str, Any]:
        """Return a dictionary to hold notebook / kernel metadata."""
        return self.json.setdefault("metadata", {})

    async def load_history(self) -> None:
        """Load kernel history."""
        await super().load_history()
        # Re-run history load for cell input-boxes
        for cell in self._rendered_cells.values():
            cell.input_box.buffer._load_history_task = None
            cell.input_box.buffer.load_history_if_not_yet_loaded()

    def kernel_started(self, result: dict[str, Any] | None = None) -> None:
        """Task to run when the kernel has started."""
        super().kernel_started(result)

    def kernel_died(self) -> None:
        """Call if the kernel dies."""
        if confirm := self.app.dialogs.get("confirm"):
            confirm.show(
                title="Kernel connection lost",
                message="The kernel appears to have died\n"
                "as it can no longer be reached.\n\n"
                "Do you want to restart the kernel?",
                cb=self.kernel.restart,
            )

    # Notebook stuff

    def load(self) -> None:
        """Load the notebook file from the file-system."""
        # Open json file, or load from passed json object
        if self.path is not None and self.path.exists():
            with self.path.open() as f:
                self.json = read_nb(f, as_version=4)
        else:
            self.json = self.json or nbformat.v4.new_notebook()
        # Ensure there is always at least one cell
        if not self.json.setdefault("cells", []):
            self.json["cells"] = [nbformat.v4.new_code_cell()]

    def set_status(self, status: str) -> None:
        """Call when kernel status changes."""
        self.cell.set_status(status)
        self.app.invalidate()

    @property
    def selected_indices(self) -> list[int]:
        """Return a list of the currently selected cell indices."""
        return []

    @abstractmethod
    def load_container(self) -> AnyContainer:
        """Absract method for loading the notebook's main container."""

    @abstractproperty
    def cell(self) -> Cell:
        """Return the current cell."""
        ...

    @property
    def current_input(self) -> KernelInput:
        """Return the currently active kernel input, if any."""
        return self.cell.input_box

    @property
    def path_name(self) -> str:
        """Return the path name."""
        if self.path is not None:
            return self.path.name
        else:
            return "(New file)"

    @property
    def title(self) -> str:
        """Return the tab title."""
        return ("* " if self.dirty else "") + self.path_name

    def lang_file_ext(self) -> str:
        """Return the file extension for scripts in the notebook's language."""
        return (
            self.json.get("metadata", {})
            .get("language_info", {})
            .get("file_extension", ".py")
        )

    def rendered_cells(self) -> list[Cell]:
        """Return a list of rendered notebooks' cells."""
        cells = {}
        for i, cell_json in enumerate(self.json.get("cells", [])):
            cell_id = get_cell_id(cell_json)
            if cell_id in self._rendered_cells:
                cells[cell_id] = self._rendered_cells[cell_id]
            else:
                cells[cell_id] = Cell(
                    i, cell_json, self, is_new=bool(self._rendered_cells)
                )
            cells[cell_id].index = i
        # These cells will be removed
        for cell in set(self._rendered_cells.values()) - set(cells.values()):
            cell.close()
            del cell
        self._rendered_cells = cells
        return list(self._rendered_cells.values())

    def get_cell_by_id(self, cell_id: str) -> Cell | None:
        """Return a reference to the `Cell` container with a given cell id."""
        # Re-render the cells as the one we want might be new
        for cell in self._rendered_cells.values():
            if cell.id == cell_id:
                break
        else:
            return None
        return cell

    def select(
        self,
        cell_index: int,
        extend: bool = False,
        position: int | None = None,
        scroll: bool = False,
    ) -> None:
        """Select a cell."""

    def scroll_to(self, index: int) -> None:
        """Scroll to a cell by index."""

    def refresh(self, slice_: slice | None = None, scroll: bool = False) -> None:
        """Refresh the notebook."""

    def refresh_cell(self, cell: Cell) -> None:
        """Trigger the refresh of a notebook cell."""

    def close(self, cb: Callable | None = None) -> None:
        """Check if the user want to save an unsaved notebook, then close the file.

        Args:
            cb: A callback to run if after closing the notebook.

        """
        if self.dirty and (unsaved := self.app.dialogs.get("unsaved")):
            unsaved.show(
                tab=self,
                cb=partial(super().close, cb),
            )
        else:
            super().close(cb)

    def write_file(self, path: Path) -> None:
        """Write the notebook's JSON to the current notebook's file.

        Additionally save the widget state to the notebook metadata.

        Args:
            path: An path at which to save the file

        """
        if self.app.config.save_widget_state:
            self.json.setdefault("metadata", {})["widgets"] = {
                "application/vnd.jupyter.widget-state+json": {
                    "version_major": 2,
                    "version_minor": 0,
                    "state": {
                        comm_id: comm._get_embed_state()
                        for comm_id, comm in self.comms.items()
                        if comm.data.get("state", {}).get("__unlinked__") is not True
                    },
                }
            }
        with path.open("w") as open_file:
            try:
                write_nb(nb=nbformat.from_dict(self.json), fp=open_file)
            except AssertionError:
                try:
                    # Jupytext requires a filename if we don't give it a format
                    write_nb(nb=nbformat.from_dict(self.json), fp=path)
                except Exception:
                    # Jupytext requires a format if the path has no extension
                    # We just use ipynb as the default format
                    write_nb(
                        nb=nbformat.from_dict(self.json),
                        fp=open_file,
                        fmt="ipynb",
                    )

    def run_cell(
        self,
        cell: Cell,
        wait: bool = False,
        callback: Callable[..., None] | None = None,
    ) -> None:
        """Run a cell.

        Args:
            cell: The rendered cell to run. If ``None``, runs the currently
                selected cell.
            wait: If :py:const:`True`, blocks until cell execution is finished
            callback: Function to run after completion

        """
        cell.remove_outputs()
        self.dirty = True
        # Queue cell if kernel not yet started
        if self.kernel.status == "starting":
            log.debug("Queuing running of cell %s", cell.index)
            self.kernel_queue.append(partial(self.run_cell, cell, wait, callback))
        else:
            self.kernel.run(
                str(cell.json.get("source")),
                wait=wait,
                callback=callback,
                get_input=cell.get_input,
                set_execution_count=cell.set_execution_count,
                add_output=cell.add_output,
                clear_output=cell.clear_output,
                set_metadata=cell.set_metadata,
                set_status=cell.set_status,
                done=cell.ran,
            )

    def new_output_default(self, output_json: dict[str, Any], own: bool) -> None:
        """Add a new output without a cell to the currently selected cell."""
        if self.app.config.show_remote_outputs:
            self.cell.add_output(output_json, own)

    def load_widgets_from_metadata(self) -> None:
        """Load widgets from state saved in notebook metadata."""
        for comm_id, comm_data in (
            self.json.get("metadata", {})
            .get("widgets", {})
            .get("application/vnd.jupyter.widget-state+json", {})
            .get("state", {})
        ).items():
            state = comm_data.get("state", {})
            # Add _model_* keys to the state
            for key, value in comm_data.items():
                if key not in ("state", "buffers"):
                    state[f"_{key}"] = value
            # Flag this widget as not linked to a widget in the kernel, so its state
            # will not be persisted on the next save
            state["__unlinked__"] = True
            # Decode base64 encoded buffers and add buffer paths to data
            buffers = []
            buffer_paths = []
            if "buffers" in state:
                for buffer_data in state.pop("buffers", []):
                    buffer_paths.append(buffer_data["path"])
                    buffers.append(standard_b64decode(buffer_data["data"]))
            # Add this comm to the notebook's current Comm map
            self.comms[comm_id] = open_comm(
                comm_container=self,
                content={
                    "data": {
                        "state": state,
                        "buffer_paths": buffer_paths,
                    },
                    "comm_id": comm_id,
                    "target_name": "jupyter.widget",
                },
                buffers=buffers,
            )

    def lsp_open_handler(self, lsp: LspClient) -> None:
        """Tell the LSP we opened a file."""
        lsp.open_nb(
            path=self.path,
            cells=[cell.lsp_cell for cell in self.rendered_cells()],
            metadata=self.metadata,
        )

    def lsp_change_handler(self, lsp: LspClient) -> None:
        """Tell the LSP server a file metadata has changed."""
        lsp.change_nb_meta(path=self.path, metadata=self.metadata)

    def lsp_before_save_handler(self, lsp: LspClient) -> None:
        """Tell the the LSP we are about to save a document."""
        # Do nothing for notebooks

    def lsp_after_save_handler(self, lsp: LspClient) -> None:
        """Tell the the LSP we saved a document."""
        lsp.save_nb(self.path)

    def lsp_close_handler(self, lsp: LspClient) -> None:
        """Tell the LSP we opened a file."""
        if lsp.can_close_nb:
            lsp.close_nb(
                path=self.path, cells=[cell.lsp_cell for cell in self.rendered_cells()]
            )
        else:
            for cell in self.rendered_cells():
                lsp.close_doc(cell.path)

    def lsp_update_diagnostics(self, lsp: LspClient) -> None:
        """Process a new diagnostic report from the LSP."""
        # Do nothing, these are handled by cells

    def __pt_searchables__(self) -> Sequence[BufferControl]:
        """Return list of cell input buffer controls for searching."""
        return [cell.input_box.control for cell in self.rendered_cells()]



"""Contain function which convert data to sixel format."""

from __future__ import annotations

from functools import partial
from typing import TYPE_CHECKING

from euporie.core.app.current import get_app
from euporie.core.convert.formats.common import (
    chafa_convert_cmd,
    chafa_convert_py,
    imagemagick_convert,
)
from euporie.core.convert.registry import register
from euporie.core.convert.utils import call_subproc
from euporie.core.filters import command_exists, have_modules

if TYPE_CHECKING:
    from typing import Any

    from euporie.core.convert.datum import Datum


register(
    from_=("png", "jpeg", "svg", "pdf"),
    to="sixel",
    filter_=command_exists("chafa"),
)(partial(chafa_convert_cmd, "sixels"))


register(
    from_=("pil"),
    to="sixel",
    filter_=have_modules("chafa"),
)(partial(chafa_convert_py, "sixels"))


@register(
    from_=("png", "jpeg"),
    to="sixel",
    filter_=command_exists("img2sixel"),
)
async def png_to_sixel_img2sixel(
    datum: Datum,
    cols: int | None = None,
    rows: int | None = None,
    fg: str | None = None,
    bg: str | None = None,
    extend: bool = True,
) -> str:
    """Convert PNG data to sixels :command:`img2sixel`."""
    cmd: list[Any] = ["img2sixel", "-I"]
    if bg:
        cmd += [f"--bgcolor={bg}"]
    if cols is not None:
        px, _ = get_app().cell_size_px
        cmd += [f"--width={int(cols * px)}"]
    return (await call_subproc(datum.data, cmd)).decode()


register(
    from_=("png", "jpeg", "svg", "pdf"),
    to="sixel",
    filter_=command_exists("magick"),
)(partial(imagemagick_convert, "sixel"))


@register(
    from_="pil",
    to="sixel",
    filter_=have_modules("timg"),
)
async def pil_to_sixel_py_timg(
    datum: Datum,
    cols: int | None = None,
    rows: int | None = None,
    fg: str | None = None,
    bg: str | None = None,
    extend: bool = True,
) -> str:
    """Convert a pillow image to sixels :py:mod:`timg`."""
    import timg

    return timg.SixelMethod(datum.data).to_string()


@register(
    from_="pil",
    to="sixel",
    filter_=have_modules("teimpy", "numpy"),
)
async def pil_to_sixel_py_teimpy(
    datum: Datum,
    cols: int | None = None,
    rows: int | None = None,
    fg: str | None = None,
    bg: str | None = None,
    extend: bool = True,
) -> str:
    """Convert a pillow image to sixels :py:mod:`teimpy`."""
    import numpy as np
    import teimpy

    return teimpy.get_drawer(teimpy.Mode.SIXEL).draw(np.asarray(datum.data))



"""Defines widget settings."""

from prompt_toolkit.filters import buffer_has_focus

from euporie.core.app.current import get_app
from euporie.core.config import add_setting

# euporie.core.widgets.cell_outputs:CellOutputArea

add_setting(
    name="wrap_cell_outputs",
    group="euporie.core.widgets.cell_outputs",
    title="wrap cell outputs",
    flags=["--wrap-cell-outputs"],
    type_=bool,
    help_="Wrap cell output text.",
    default=False,
    schema={"type": "boolean"},
    description="""
        Whether text-based cell outputs should be wrapped.
    """,
    cmd_filter=~buffer_has_focus,
)

add_setting(
    name="text_output_limit",
    group="euporie.core.widgets.cell_outputs",
    flags=["--text-output-limit"],
    type_=int,
    help_="Limit the amount of cell text output",
    default=1_000_000,
    schema={
        "minimum": 0,
    },
    description="""
        Limit the number of text characters in interactive cell text output to this value.
        Use ``0`` to allow any amount of characters.
    """,
)

# euporie,core.widgets.file_browser:FileBrowser

add_setting(
    name="show_file_icons",
    group="euporie.core.widgets.file_browser",
    flags=["--show-file-icons"],
    type_=bool,
    title="File icons",
    help_="Show file icons in the file manager",
    default=False,
    schema={
        "type": "boolean",
    },
    description="""
        Whether file icons should be shown in the file manager.

        These icons exist in the unicode private use area, and may require custom
        fonts such as ``awesome-terminal-fonts`` or ``nerdfonts`` to be installed.
    """,
)

# euporie.core.widgets.inputs:KernelInput

add_setting(
    name="line_numbers",
    group="euporie.core.widgets.inputs",
    flags=["--line-numbers"],
    type_=bool,
    help_="Show or hide line numbers",
    default=True,
    description="""
        Whether line numbers are shown by default.
    """,
    hooks=[lambda x: get_app().refresh()],
)

add_setting(
    name="autoformat",
    group="euporie.core.widgets.inputs",
    flags=["--autoformat"],
    type_=bool,
    help_="Automatically re-format code cells when run",
    default=False,
    description="""
        Whether to automatically reformat code cells before they are run.
    """,
)

add_setting(
    name="autocomplete",
    group="euporie.core.widgets.inputs",
    flags=["--autocomplete"],
    type_=bool,
    help_="Provide completions suggestions automatically",
    default=False,
    description="""
        Whether to automatically suggestion completions while typing in code cells.
    """,
)

add_setting(
    name="autosuggest",
    group="euporie.core.widgets.inputs",
    flags=["--autosuggest"],
    type_=str,
    choices=["smart", "simple", "none"],
    help_="Provide line completion suggestions",
    default="smart",
    description="""
        Whether to automatically suggestion line content while typing in code cells.
    """,
)

add_setting(
    name="autoinspect",
    group="euporie.core.widgets.inputs",
    flags=["--autoinspect"],
    type_=bool,
    help_="Display contextual help automatically",
    default=False,
    description="""
        Whether to automatically display contextual help when navigating through code cells.
    """,
)


# euporie.core.bars.status:StatusBar

add_setting(
    name="show_status_bar",
    group="euporie.core.bars.status",
    flags=["--show-status-bar"],
    type_=bool,
    title="status bar",
    help_="Show the status bar",
    default=True,
    schema={
        "type": "boolean",
    },
    description="""
        Whether the status bar should be shown at the bottom of the screen.
    """,
)

# euporie.core.widgets.decor:

add_setting(
    name="show_shadows",
    group="euporie.core.widgets.decor",
    flags=["--show-shadows"],
    type_=bool,
    help_="Show or hide shadows under menus and dialogs",
    default=True,
    description="""
        Sets whether shadows are shown under dialogs and popup-menus.
    """,
)

# euporie.core.widgets.cell

add_setting(
    name="show_cell_borders",
    group="euporie.core.widgets.cell",
    title="cell borders",
    flags=["--show-cell-borders"],
    type_=bool,
    help_="Show or hide cell borders.",
    default=False,
    schema={
        "type": "boolean",
    },
    description="""
        Whether cell borders should be drawn for unselected cells.
    """,
)

add_setting(
    name="external_editor",
    group="euporie.core.widgets.cell",
    flags=["--external-editor"],
    type_=str,
    help_="Set the external editor to use.",
    default=None,
    description="""
        A command to run when editing cells externally. The following strings in
        the command will be replaced with values which locate the cell being
        edited:

        * ``{top}``
        * ``{left}``
        * ``{bottom}``
        * ``{right}``
        * ``{width}``
        * ``{height}``

        This is useful if you run euporie inside a tmux session, and wish to launch
        your editor in a pop-up pane. This can be achieved by setting this parameter
        to something like the following:

        .. code-block::

           "tmux display-popup -x {left} -y {bottom} -w {width} -h {height} -B -E micro"

    """,
)



"""Contains containers which display children at full height vertically stacked."""

from __future__ import annotations

import asyncio
import logging
from typing import TYPE_CHECKING, cast

from prompt_toolkit.application.current import get_app
from prompt_toolkit.cache import FastDictCache
from prompt_toolkit.filters import is_searching
from prompt_toolkit.layout.containers import (
    Container,
    ScrollOffsets,
    Window,
    WindowRenderInfo,
)
from prompt_toolkit.layout.controls import UIContent
from prompt_toolkit.layout.dimension import Dimension, to_dimension
from prompt_toolkit.mouse_events import MouseEvent, MouseEventType, MouseModifier

from euporie.core.layout.cache import CachedContainer
from euporie.core.layout.screen import BoundedWritePosition

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import Callable, Literal

    from prompt_toolkit.key_binding.key_bindings import NotImplementedOrNone
    from prompt_toolkit.layout.containers import AnyContainer
    from prompt_toolkit.layout.dimension import AnyDimension
    from prompt_toolkit.layout.mouse_handlers import MouseHandlers
    from prompt_toolkit.layout.screen import Screen as PtkScreen
    from prompt_toolkit.layout.screen import WritePosition

    MouseHandler = Callable[[MouseEvent], object]

log = logging.getLogger(__name__)


class ScrollingContainer(Container):
    """A scrollable container which renders only the currently visible children."""

    render_info: WindowRenderInfo | None

    def __init__(
        self,
        children: Callable[[], Sequence[AnyContainer]] | Sequence[AnyContainer],
        height: AnyDimension = None,
        width: AnyDimension = None,
        style: str | Callable[[], str] = "",
        scroll_offsets: ScrollOffsets | None = None,
    ) -> None:
        """Initiate the `ScrollingContainer`."""
        if callable(children):
            _children_func = children
        else:

            def _children_func() -> Sequence[AnyContainer]:
                """Return the pass sequence of children."""
                assert not callable(children)
                return children

        self.children_func = _children_func
        self._child_cache: dict[int, CachedContainer] = {}
        self._children: list[CachedContainer] = []
        self._known_sizes_cache: FastDictCache[tuple[int], list[int]] = FastDictCache(
            self._known_sizes
        )
        self.refresh_children = True
        self.pre_rendered: float | None = None

        self._selected_slice = slice(
            0, 1
        )  # The index of the currently selected children
        self.selected_child_position: int = 0

        self.visible_indices: set[int] = {0}
        self.index_positions: dict[int, int | None] = {}

        self.last_write_position: WritePosition = BoundedWritePosition(0, 0, 0, 0)
        self.last_total_height = 0

        self.width = to_dimension(width).preferred
        self.height = to_dimension(height).preferred
        self.style = style
        self.scroll_offsets = scroll_offsets or ScrollOffsets(
            top=1, bottom=1, left=1, right=1
        )

        self.scroll_to_cursor = False
        self.scrolling = 0

    def pre_render_children(self, width: int, height: int) -> None:
        """Render all unrendered children in a background thread."""
        children = self.all_children()
        if not children:
            return
        self.pre_rendered = 0.0
        incr = 1 / len(children)
        app = get_app()

        def _cb(task: asyncio.Task) -> None:
            """Task callback to update pre-rendering percentage."""
            assert isinstance(self.pre_rendered, float)
            self.pre_rendered += incr
            app.invalidate()

        tasks = set()
        for child in children:
            if isinstance(child, CachedContainer):
                task = app.create_background_task(
                    asyncio.to_thread(child.render, width, height)
                )
                task.add_done_callback(_cb)
                tasks.add(task)

        async def _finish() -> None:
            await asyncio.gather(*tasks)
            self.pre_rendered = 1.0
            app.invalidate()
            # app.exit()

        app.create_background_task(_finish())

    def reset(self) -> None:
        """Reset the state of this container and all the children."""
        for child in self.all_children():
            child.reset()

    def preferred_width(self, max_available_width: int) -> Dimension:
        """Do not provide a preferred width - grow to fill the available space."""
        if self.width is not None:
            return Dimension(min=1, preferred=self.width, weight=1)
        return Dimension(weight=1)

    def preferred_height(self, width: int, max_available_height: int) -> Dimension:
        """Return the preferred height only if one is provided."""
        if self.height is not None:
            return Dimension(min=1, preferred=self.height, weight=1)
        return Dimension(weight=1)

    def _set_selected_slice(
        self, req_slice: slice, force: bool = False, scroll: bool = True
    ) -> None:
        # Only update the selected child if it was not selected before
        if force or req_slice != self._selected_slice:
            app = get_app()
            self.refresh_children = True
            children = self.all_children()
            # Ensure new selected slice is valid
            new_slice = self.validate_slice(req_slice)
            # Scroll into view
            if scroll:
                anchor: Literal["top", "bottom"] | None = None
                if (
                    new_slice.start == len(children) - 1
                    and new_slice.stop
                    and req_slice.stop > new_slice.stop
                ):
                    anchor = "bottom"
                elif req_slice.start == -1:
                    anchor = "top"
                self.scroll_to(new_slice.start, anchor)
            # Request a refresh of the previously selected children
            for child in self._selected_children:
                child.invalidate()
            # Get the first selected child and focus it
            child = children[new_slice.start]
            if not app.layout.has_focus(child):
                try:
                    app.layout.focus(child)
                except ValueError:
                    pass
            # Track which child was selected
            self._selected_slice = new_slice

    @property
    def selected_slice(self) -> slice:
        """Return the currently selected slice."""
        return self._selected_slice

    @selected_slice.setter
    def selected_slice(self, new_slice: slice) -> None:
        """Set the currently selected child index.

        Args:
            new_slice: The slice of the children to select.
        """
        self._set_selected_slice(new_slice)

    def validate_slice(self, slice_: slice) -> slice:
        """Ensure a slice describes a valid range of children."""
        start = min(max(slice_.start, 0), len(self._children) - 1)
        stop = slice_.stop
        if stop == -1:
            stop = None
        if stop is not None:
            stop = min(max(slice_.stop, int(start == 0)), len(self._children))
        return slice(start, stop, slice_.step)

    @property
    def selected_indices(self) -> list[int]:
        """Return in indices of the currently selected children."""
        return list(range(*self.selected_slice.indices(len(self._children))))

    def select(
        self,
        index: int,
        extend: bool = False,
        position: int | None = None,
        scroll: bool = True,
    ) -> None:
        """Select a child or adds it to the selection.

        Args:
            index: The index of the cell to select
            extend: If true, the selection will be extended to include the cell
            position: An optional cursor position index to apply to the cell input
            scroll: Whether to scroll the page
        """
        # Update the selected slice if we are extending the child selection
        if extend:
            slice_ = self._selected_slice
            stop = -1 if slice_.stop is None else slice_.stop
            step = slice_.step
            if step == -1 and index <= stop:
                stop += 2
                step = 1
            elif step == -1 and index >= stop:
                pass
            elif step in (1, None) and index < stop:
                step = 1
            elif step in (1, None) and index >= stop:
                step = -1
                stop -= 2
            self._set_selected_slice(slice(index, stop, step), scroll=scroll)
        # Otherwise set the cell selection to the given cell index
        else:
            self._set_selected_slice(slice(index, index + 1), scroll=scroll)

    def scroll(self, n: int) -> NotImplementedOrNone:
        """Scroll up or down a number of rows.

        Args:
            n: The number of rows to scroll, negative for up, positive for down

        Returns:
            :py:const:`NotImplemented` is scrolling is not allowed, otherwise
                :py:const:`None`

        """
        if n > 0:
            if (
                min(self.visible_indices) == 0
                and self.index_positions
                and self.index_positions[0] is not None
            ):
                n = min(n, 0 - self.index_positions[0] - self.scrolling)
                if self.index_positions[0] + self.scrolling + n > 0:
                    return NotImplemented
        elif n < 0:
            bottom_index = len(self._children) - 1
            if bottom_index in self.visible_indices:
                bottom_child = self.get_child(bottom_index)
                bottom_pos = self.index_positions[bottom_index]
                if bottom_pos is not None:
                    n = max(
                        n,
                        self.last_write_position.height
                        - (bottom_pos + bottom_child.height + self.scrolling),
                    )
                    if (
                        bottom_pos + bottom_child.height + self.scrolling + n
                        < self.last_write_position.height
                    ):
                        return NotImplemented

        # Very basic scrolling acceleration
        if n:
            self.scrolling += n
            return None
        else:
            return NotImplemented

    def mouse_scroll_handler(self, mouse_event: MouseEvent) -> NotImplementedOrNone:
        """Mouse handler to scroll the pane."""
        if mouse_event.event_type == MouseEventType.SCROLL_DOWN:
            return self.scroll(-1)
        elif mouse_event.event_type == MouseEventType.SCROLL_UP:
            return self.scroll(1)
        else:
            return NotImplemented
        return None

    def _mouse_handler_wrapper(
        self, handler: MouseHandler | None = None, child: CachedContainer | None = None
    ) -> MouseHandler:
        def wrapped(mouse_event: MouseEvent) -> NotImplementedOrNone:
            response: NotImplementedOrNone = NotImplemented

            if callable(handler):
                response = handler(mouse_event)

            if response is NotImplemented:
                if mouse_event.event_type == MouseEventType.SCROLL_DOWN:
                    response = self.scroll(-1)
                elif mouse_event.event_type == MouseEventType.SCROLL_UP:
                    response = self.scroll(1)

            # Refresh the child if there was a response
            if response is None:
                return response

            # Select the clicked child if clicked
            if mouse_event.event_type == MouseEventType.MOUSE_DOWN:
                if child:
                    index = self._children.index(child)
                    if mouse_event.modifiers & {
                        MouseModifier.SHIFT,
                        MouseModifier.CONTROL,
                    }:
                        self.select(index, extend=True)
                    else:
                        self.select(index, extend=False)
                    get_app().layout.focus(child)
                else:
                    try:
                        get_app().layout.focus(self)
                    except Exception:
                        ...
                response = None

            return response

        return wrapped

    def write_to_screen(
        self,
        screen: PtkScreen,
        mouse_handlers: MouseHandlers,
        write_position: WritePosition,
        parent_style: str,
        erase_bg: bool,
        z_index: int | None,
    ) -> None:
        """Write the actual content to the screen.

        Children are rendered only if they are visible and have changed, and the output
        is cached to a separate screen object stored in a :py::class:`ChildRenderInfo`. The
        cached rendering of the children which are actually visible are then copied to
        the screen. This results in faster rendering of the scrolling container, and
        makes scrolling more performant.

        Args:
            screen: The :class:`~prompt_toolkit.layout.screen.Screen` class to which
                the output has to be written.
            mouse_handlers: :class:`prompt_toolkit.layout.mouse_handlers.MouseHandlers`.
            write_position: A :class:`prompt_toolkit.layout.screen.WritePosition` object
                defining where this container should be drawn.
            erase_bg: If true, the background will be erased prior to drawing.
            parent_style: Style string to pass to the :class:`.Window` object. This will
                be applied to all content of the windows. :class:`.VSplit` and
                :class:`prompt_toolkit.layout.containers.HSplit` can use it to pass
                their style down to the windows that they contain.
            z_index: Used for propagating z_index from parent to child.
        """
        # Record where the container was last drawn so we can determine if cell outputs
        # are partially obscured
        self.last_write_position = write_position

        ypos = write_position.ypos
        xpos = write_position.xpos

        available_width = write_position.width
        available_height = write_position.height

        # Update screen height
        screen.height = max(screen.height, ypos + write_position.height)

        # Record children which are currently visible
        visible_indices = set()

        # Force the selected children to refresh
        selected_indices = self.selected_indices
        self._selected_children: list[CachedContainer] = []
        for index in selected_indices:
            child = self.get_child(index)
            # Do not bother to re-render selected children if we are scrolling
            if not self.scrolling:
                child.invalidate()
            self._selected_children.append(child)
            self.index_positions[index] = None

        # Refresh **visible** children if searching
        if is_searching():
            for index in self.visible_indices:
                self.get_child(index).invalidate()

        # Scroll to make the cursor visible
        layout = get_app().layout
        if self.scroll_to_cursor:
            selected_child = self._selected_children[0]
            selected_child.render(
                available_width=available_width,
                available_height=available_height,
                style=f"{parent_style} {self.style}",
                start=self.selected_child_position,
            )
            if cursor_position := selected_child.screen.cursor_positions.get(
                layout.current_window
            ):
                cursor_row = self.selected_child_position + cursor_position.y
                scroll_offsets = self.scroll_offsets
                if cursor_row < scroll_offsets.top + scroll_offsets.top:
                    self.selected_child_position -= cursor_row - scroll_offsets.top
                elif cursor_row >= available_height - (scroll_offsets.bottom + 1):
                    self.selected_child_position -= (
                        cursor_row - available_height + (scroll_offsets.bottom + 1)
                    )
            # Set this to false again, allowing us to scroll the cursor out of view
            self.scroll_to_cursor = False

        # Adjust scrolling offset
        heights = self.known_sizes
        total_height = sum(heights)
        if self.scrolling or total_height != self.last_total_height:
            heights_above = sum(heights[: self._selected_slice.start])
            new_child_position = self.selected_child_position + self.scrolling
            # Do not allow scrolling if there is no overflow
            if sum(heights) < available_height:
                self.selected_child_position = heights_above
                self.scrolling = 0
            else:
                # Prevent overscrolling at the top of the document
                overscroll = heights_above - new_child_position
                if overscroll < 0:
                    # Scroll as far as we are able without overscrolling
                    self.scrolling = max(0, self.scrolling + overscroll)
                # Prevent underscrolling at the bottom
                elif overscroll > 0:
                    heights_below = sum(heights[self._selected_slice.start :])
                    underscroll = new_child_position + heights_below - available_height
                    if underscroll < 0:
                        # Scroll as far as we are able without underscrolling
                        self.scrolling = min(0, self.scrolling - underscroll)
            self.selected_child_position += self.scrolling

        self.last_total_height = total_height

        # Blit first selected child and those below it that are on screen
        line = self.selected_child_position
        for i in range(self._selected_slice.start, len(self._children)):
            child = self.get_child(i)
            child.render(
                available_width=available_width,
                available_height=available_height,
                style=f"{parent_style} {self.style}",
                start=line,
            )
            if line + child.height > 0 and line < available_height:
                self.index_positions[i] = line
                child.blit(
                    screen=screen,
                    mouse_handlers=mouse_handlers,
                    left=xpos,
                    top=ypos + line,
                    cols=slice(0, available_width),
                    rows=slice(
                        max(0, 0 - line),
                        min(child.height, available_height - line),
                    ),
                )
                visible_indices.add(i)
            line += child.height
            if line >= available_height:
                break
        else:
            if line < available_height:
                Window(char=" ").write_to_screen(
                    screen,
                    mouse_handlers,
                    BoundedWritePosition(
                        xpos, ypos + line, available_width, available_height - line
                    ),
                    parent_style,
                    erase_bg,
                    z_index,
                )
                for y in range(ypos + line, ypos + available_height):
                    for x in range(xpos, xpos + available_width):
                        mouse_handlers.mouse_handlers[y][x] = (
                            self._mouse_handler_wrapper(
                                mouse_handlers.mouse_handlers[y][x]
                            )
                        )
        # Blit children above the selected that are on screen
        line = self.selected_child_position
        for i in range(self._selected_slice.start - 1, -1, -1):
            child = self.get_child(i)
            child.render(
                available_width=available_width,
                available_height=available_height,
                style=f"{parent_style} {self.style}",
                end=line,
            )
            # TODOD - prevent lagged child height
            line -= child.height
            if line + child.height > 0 and line < available_height:
                self.index_positions[i] = line
                child.blit(
                    screen=screen,
                    mouse_handlers=mouse_handlers,
                    left=xpos,
                    top=ypos + line,
                    cols=slice(0, available_width),
                    rows=slice(
                        max(0, 0 - line),
                        min(child.height, available_height - line),
                    ),
                )
                visible_indices.add(i)
            if line <= 0:
                break
        else:
            if line > 0:
                Window(char=" ").write_to_screen(
                    screen,
                    mouse_handlers,
                    BoundedWritePosition(xpos, ypos, available_width, line),
                    parent_style,
                    erase_bg,
                    z_index,
                )
                for y in range(ypos, ypos + line):
                    for x in range(xpos, xpos + available_width):
                        mouse_handlers.mouse_handlers[y][x] = (
                            self._mouse_handler_wrapper(
                                mouse_handlers.mouse_handlers[y][x]
                            )
                        )

        # Dont bother drawing floats
        # screen.draw_all_floats()

        # Ensure the focused child is always in the layout
        visible_indices.add(self._selected_slice.start)

        # Update which children will appear in the layout
        self.visible_indices = visible_indices

        # Update parent relations in layout
        def _walk(e: Container) -> None:
            for c in e.get_children():
                layout._child_to_parent[c] = e
                _walk(c)

        _walk(self)

        # Mock up a WindowRenderInfo so we can draw a scrollbar margin
        self.render_info = WindowRenderInfo(
            window=cast("Window", self),
            ui_content=UIContent(line_count=max(sum(heights), 1)),
            horizontal_scroll=0,
            vertical_scroll=self.vertical_scroll,
            window_width=available_width,
            window_height=available_height,
            configured_scroll_offsets=ScrollOffsets(),
            visible_line_to_row_col={i: (i, 0) for i in range(available_height)},
            rowcol_to_yx={},
            x_offset=xpos,
            y_offset=ypos,
            wrap_lines=False,
        )
        # Signal that we are no longer scrolling
        self.scrolling = 0

        # Trigger pre-rendering of children
        if self.pre_rendered is None:
            self.pre_render_children(available_width, available_height)

    @property
    def known_sizes(self) -> list[int]:
        """Map of child indices to height values."""
        return self._known_sizes_cache[get_app().render_counter,]

    def _known_sizes(self, render_counter: int) -> list[int]:
        """Calculate sizes of children once per render cycle."""
        sizes = {}
        missing = set()
        available_width = self.last_write_position.width
        available_height = self.last_write_position.height
        for i, child in enumerate(self._children):
            if isinstance(child, CachedContainer) and child.height:
                sizes[i] = child.preferred_height(
                    available_width, available_height
                ).preferred
            else:
                missing.add(i)
        avg = int(sum(sizes.values()) / (len(sizes) or 1))
        sizes.update(dict.fromkeys(missing, avg))
        return [v for k, v in sorted(sizes.items())]

    @property
    def vertical_scroll(self) -> int:
        """The best guess at the absolute vertical scroll position."""
        return (
            sum(self.known_sizes[: self._selected_slice.start])
            - self.selected_child_position
        )

    @vertical_scroll.setter
    def vertical_scroll(self, value: int) -> None:
        """Set the absolute vertical scroll position."""
        self.selected_child_position = (
            sum(self.known_sizes[: self._selected_slice.start]) - value
        )

    def all_children(self) -> Sequence[Container]:
        """Return the list of all children of this container."""
        _children = self._children
        if self.refresh_children or not self._children:
            self.refresh_children = False
            new_children = []
            new_child_hashes = set()
            for child in self.children_func():
                if not (
                    wrapped_child := (self._child_cache.get(child_hash := hash(child)))
                ):
                    wrapped_child = self._child_cache[child_hash] = CachedContainer(
                        child, mouse_handler_wrapper=self._mouse_handler_wrapper
                    )
                new_children.append(wrapped_child)
                new_child_hashes.add(child_hash)
            _children[:] = new_children

            # Clean up metacache
            for child_hash in set(self._child_cache) - new_child_hashes:
                del self._child_cache[child_hash]

            # Clean up positions
            self.index_positions = {
                i: pos
                for i, pos in self.index_positions.items()
                if i < len(self._children)
            }
        return _children

    def get_children(self) -> list[Container]:
        """Return the list of currently visible children to include in the layout."""
        # Ensure children are loaded
        self.all_children()
        # Return only the visible children
        return [
            self.get_child(i) for i in self.visible_indices if i < len(self._children)
        ]

    def get_child(self, index: int | None = None) -> CachedContainer:
        """Return a rendered instance of the child at the given index.

        If no index is given, the currently selected child is returned.

        Args:
            index: The index of the child to return.

        Returns:
            A rendered instance of the child.

        """
        if index is None:
            index = self.selected_slice.start
        if self._children:
            index = max(0, min(len(self._children) - 1, index))
            return self._children[index]
        else:
            return CachedContainer(Window())

    def scroll_to(
        self, index: int, anchor: Literal["top", "bottom"] | None = None
    ) -> None:
        """Scroll a child into view.

        Args:
            index: The child index to scroll into view
            anchor: Whether to scroll to the top or bottom the given child index

        """
        child = self.get_child(index)

        new_top: int | None = None
        if index in self.visible_indices:
            new_top = self.index_positions[index]
        else:
            if index < self._selected_slice.start:
                new_top = max(0, child.height - self.last_write_position.height)
            elif index == self._selected_slice.start:
                new_top = self.selected_child_position
            elif index > self._selected_slice.start:
                indices = [k for k, v in self.index_positions.items() if v is not None]
                if indices:
                    last_index = max(indices)
                    new_top = max(
                        min(
                            self.last_write_position.height,
                            (self.index_positions[last_index] or 0)
                            + self.get_child(last_index).height,
                        ),
                        0,
                    )

        # If the child height is not know, we'll need to render it to determine its height
        if not child.height:
            child.render(
                self.last_write_position.width,
                self.last_write_position.height,
            )

        if new_top is None or new_top < 0:
            if anchor == "top":
                self.selected_child_position = 0
            else:
                self.selected_child_position = min(
                    0, self.last_write_position.height - child.height
                )
        elif new_top > self.last_write_position.height - child.height:
            self.selected_child_position = max(
                0, self.last_write_position.height - child.height
            )
            if anchor == "bottom":
                self.selected_child_position -= (
                    child.height - self.last_write_position.height
                )
        else:
            self.selected_child_position = new_top

    def _scroll_up(self) -> None:
        """Scroll up one line: for compatibility with :py:class:`Window`."""
        self.scroll(1)

    def _scroll_down(self) -> None:
        """Scroll down one line: for compatibility with :py:class:`Window`."""
        self.scroll(-1)



"""Contain classes relating to kernel management."""

from __future__ import annotations

import logging
import os
import re
import sys
import threading
from typing import TYPE_CHECKING

from jupyter_client import AsyncKernelManager
from jupyter_client.provisioning.local_provisioner import LocalProvisioner

if TYPE_CHECKING:
    from typing import Any, Callable, TextIO

    from jupyter_client.connect import KernelConnectionInfo


log = logging.getLogger(__name__)


class LoggingLocalProvisioner(LocalProvisioner):  # type:ignore[misc]
    """A Jupyter kernel provisionser which logs kernel output."""

    async def launch_kernel(
        self, cmd: list[str], **kwargs: Any
    ) -> KernelConnectionInfo:
        """Launch a kernel with a command."""
        await super().launch_kernel(cmd, **kwargs)

        def log_kernel_output(pipe: TextIO, log_func: Callable) -> None:
            try:
                if pipe:
                    with pipe:
                        for line in iter(pipe.readline, ""):
                            log_func(line.rstrip())
            except StopIteration:
                pass

        if self.process is not None:
            # Start thread to listen for kernel output
            threading.Thread(
                target=log_kernel_output,
                args=(self.process.stdout, log.warning),
                daemon=True,
            ).start()

        return self.connection_info


def set_default_provisioner() -> None:
    """Set the default kernel provisioner to euporie's logging provisioner."""
    from jupyter_client.provisioning import KernelProvisionerFactory as KPF

    KPF.instance().default_provisioner_name = "logging-local-provisioner"


class EuporieKernelManager(AsyncKernelManager):
    """Kernel Manager subclass.

    ``jupyter_client`` replaces a plain ``python`` command with the current executable,
    but this is not desirable if the client is running in its own prefix (e.g. with
    ``pipx``). We work around this here.

    See https://github.com/jupyter/jupyter_client/issues/949
    """

    def format_kernel_cmd(self, extra_arguments: list[str] | None = None) -> list[str]:
        """Replace templated args (e.g. {connection_file})."""
        extra_arguments = extra_arguments or []
        assert self.kernel_spec is not None
        cmd = self.kernel_spec.argv + extra_arguments

        v_major, v_minor = sys.version_info[:2]
        if cmd and cmd[0] in {
            "python",
            f"python{v_major}",
            f"python{v_major}.{v_minor}",
        }:
            # If the command is `python` without an absolute path and euporie is
            # running in the same prefix as the kernel_spec file is located, use
            # sys.executable: otherwise fall back to the executable in the base prefix
            if (
                os.path.commonpath((sys.prefix, self.kernel_spec.resource_dir))
                == sys.prefix
            ):
                cmd[0] = sys.executable
            else:
                cmd[0] = sys._base_executable  # type: ignore [attr-defined]

        # Make sure to use the realpath for the connection_file
        # On windows, when running with the store python, the connection_file path
        # is not usable by non python kernels because the path is being rerouted when
        # inside of a store app.
        # See this bug here: https://bugs.python.org/issue41196
        ns = {
            "connection_file": os.path.realpath(self.connection_file),
            "prefix": sys.prefix,
        }

        if self.kernel_spec:
            ns["resource_dir"] = self.kernel_spec.resource_dir

        if self._launch_args:
            ns.update({str(k): str(v) for k, v in self._launch_args.items()})

        pat = re.compile(r"\{([A-Za-z0-9_]+)\}")

        def _from_ns(match: re.Match) -> str:
            """Get the key out of ns if it's there, otherwise no change."""
            return ns.get(match.group(1), match.group())

        return [pat.sub(_from_ns, arg) for arg in cmd]



"""Contain ANSI formatted text parser."""

from __future__ import annotations

import logging
import re
from typing import TYPE_CHECKING

from prompt_toolkit.formatted_text import ANSI as PTANSI

if TYPE_CHECKING:
    from collections.abc import Generator

log = logging.getLogger(__name__)


class ANSI(PTANSI):
    """Convert ANSI text into formatted text, preserving all control sequences."""

    def __init__(self, value: str, tab_size: int = 8) -> None:
        """Initiate the ANSI processor instance.

        This replaces carriage returns to emulate terminal output.

        Args:
            value: The ANSI string to process.
            tab_size: The number of spaces to use to represent a tab

        """
        # Replace tabs with spaces
        value = value.expandtabs(tabsize=tab_size)
        # Replace windows style newlines
        value = value.replace("\r\n", "\n")
        # Remove anything before a carriage return if there is something after it to
        # emulate a carriage return in the output
        value = re.sub(r"^.*\r(?!\n)", "", value, count=0, flags=re.MULTILINE)
        # Clear line by deleting previous characters
        value = re.sub(r".*\x1b\[2K", "", value, count=0)
        # Remove hide & show cursor commands
        value = re.sub(r"\x1b\[\?25[hl]", "", value, count=0)

        super().__init__(value)

    def _parse_corot(self) -> Generator[None, str, None]:
        """Coroutine that parses the ANSI escape sequences.

        This is modified version of the ANSI parser from prompt_toolkit retains
        all CSI escape sequences.

        Yields:
            Accepts characters from a string.

        """
        style = ""
        formatted_text = self._formatted_text

        while True:
            char = yield
            sequence = char

            # Everything between \001 and \002 should become a ZeroWidthEscape.
            if char == "\001":
                sequence = ""
                while char != "\002":
                    char = yield
                    if char == "\002":
                        formatted_text.append(("[ZeroWidthEscape]", sequence))
                        break
                    else:
                        sequence += char
                continue

            # Check for backspace
            elif char == "\x08":
                # TODO - remove last character from last non-ZeroWidthEscape fragment
                if formatted_text:
                    formatted_text.pop()
                continue

            elif char in ("\x1b", "\x9b"):
                # Got a CSI sequence, try to compile a control sequence
                char = yield

                # Check for sixels
                if char == "P":
                    # Got as DEC code
                    sequence += char
                    # We expect "p1;p2;p3;q" + sixel data + "\x1b\"
                    char = yield
                    while char != "\x1b":
                        sequence += char
                        char = yield
                    sequence += char
                    char = yield
                    if ord(char) == 0x5C:
                        sequence += char
                        formatted_text.append(("[ZeroWidthEscape]", sequence))
                        # char = yield
                        continue

                # Check for hyperlinks
                elif char == "]":
                    sequence += char
                    char = yield
                    if char == "8":
                        sequence += char
                        char = yield
                        if char == ";":
                            sequence += char
                            char = yield
                            while True:
                                sequence += char
                                if sequence[-2:] == "\x1b\\":
                                    break
                                char = yield
                            formatted_text.append(("[ZeroWidthEscape]", sequence))
                            continue

                elif (char == "[" and sequence == "\x1b") or sequence == "\x9b":
                    if sequence == "\x1b":
                        sequence += char
                        char = yield

                    # Next are any number (including none) of "parameter bytes"
                    params = []
                    current = ""
                    while 0x30 <= ord(char) <= 0x3F:
                        # Parse list of integer parameters
                        sequence += char
                        if char.isdigit():
                            current += char
                        else:
                            params.append(min(int(current or 0), 9999))
                            if char == ";":
                                current = ""
                        char = yield
                    if current:
                        params.append(min(int(current or 0), 9999))
                    # then any number of "intermediate bytes"
                    while 0x20 <= ord(char) <= 0x2F:
                        sequence += char
                        char = yield
                    # finally by a single "final byte"
                    if 0x40 <= ord(char) <= 0x7E:
                        sequence += char
                    # Check if that escape sequence was a style:
                    if char == "m":
                        self._select_graphic_rendition(params)
                        style = self._create_style_string()
                    # Otherwise print a zero-width control sequence
                    else:
                        formatted_text.append(("[ZeroWidthEscape]", sequence))
                    continue

            formatted_text.append((style, sequence))



"""Contains commands common to all euporie applications."""

from __future__ import annotations

from typing import TYPE_CHECKING

from prompt_toolkit.filters import buffer_has_focus

from euporie.core.app.current import get_app
from euporie.core.commands import add_cmd
from euporie.core.filters import tab_has_focus, tab_type_has_focus

if TYPE_CHECKING:
    from prompt_toolkit.key_binding.key_processor import KeyPressEvent


@add_cmd(aliases=["q"])
def _quit() -> None:
    """Quit euporie."""
    get_app().exit()


@add_cmd(aliases=["q!"])
def _force_quit() -> None:
    """Quit euporie without saving any changes."""
    from prompt_toolkit.application.application import Application

    Application.exit(get_app())


@add_cmd(aliases=["bc"], filter=tab_has_focus, menu_title="Close File")
def _close_tab() -> None:
    """Close the current tab."""
    get_app().close_tab()


@add_cmd(aliases=["bn"], filter=tab_has_focus)
def _next_tab() -> None:
    """Switch to the next tab."""
    get_app().tab_idx += 1


@add_cmd(aliases=["bp"], filter=tab_has_focus)
def _previous_tab() -> None:
    """Switch to the previous tab."""
    get_app().tab_idx -= 1


@add_cmd(filter=~buffer_has_focus)
def _focus_next() -> None:
    """Focus the next control."""
    get_app().layout.focus_next()


@add_cmd(filter=~buffer_has_focus)
def _focus_previous() -> None:
    """Focus the previous control."""
    get_app().layout.focus_previous()


@add_cmd()
def _clear_screen() -> None:
    """Clear the screen."""
    get_app().renderer.clear()


@add_cmd(hidden=True, aliases=[""])
def _go_to(event: KeyPressEvent, index: int = 0) -> None:
    """Go to a line or cell by number."""
    index = max(0, index - 1)
    if buffer_has_focus():
        buffer = get_app().current_buffer
        buffer.cursor_position = len("".join(buffer.text.splitlines(True)[:index]))
    elif tab_type_has_focus("euporie.notebook.tabs.notebook:Notebook")():
        from euporie.notebook.tabs.notebook import Notebook

        if isinstance(nb := get_app().tab, Notebook):
            nb.select(index)



"""Show contextual help for the current item under the cursor."""

from __future__ import annotations

import asyncio
from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Sequence
    from pathlib import Path
    from typing import Any, Callable

    from prompt_toolkit.document import Document

    from euporie.core.kernel.base import BaseKernel
    from euporie.core.lsp import LspClient


class Inspector(metaclass=ABCMeta):
    """A class which provides contextual help on the item under the cursor."""

    @abstractmethod
    async def get_context(self, document: Document, auto: bool) -> dict[str, Any]:
        """Call to retrieve the contextual help."""
        ...


class KernelInspector(Inspector):
    """Inspector which retrieves contextual help from a kernel."""

    def __init__(self, kernel: BaseKernel | Callable[[], BaseKernel]) -> None:
        """Initialize a new inspector which queries a kernel."""
        self._kernel = kernel

    @property
    def kernel(self) -> BaseKernel:
        """Return the current kernel."""
        if callable(self._kernel):
            return self._kernel()
        return self._kernel

    async def get_context(self, document: Document, auto: bool) -> dict[str, Any]:
        """Request contextual help from the kernel."""
        return await self.kernel.inspect_async(document.text, document.cursor_position)


class LspInspector(Inspector):
    """Inspector which retrieves contextual help from a Language Server."""

    def __init__(self, lsp: LspClient, path: Path) -> None:
        """Initialize a new inspector which queries a kernel."""
        self.lsp = lsp
        self.path = path

    async def get_context(self, document: Document, auto: bool) -> dict[str, Any]:
        """Request hover info from an LSP servers."""
        return await self.lsp.hover_(
            path=self.path,
            line=document.cursor_position_row,
            char=document.cursor_position_col,
        )


class FirstInspector(Inspector):
    """Return results of the first inspector to respond."""

    def __init__(
        self, inspectors: Sequence[Inspector] | Callable[[], Sequence[Inspector]]
    ) -> None:
        """Initialize a new inspector which queries a kernel."""
        self.inspectors = inspectors

    async def get_context(self, document: Document, auto: bool) -> dict[str, Any]:
        """Request hover info from an LSP servers."""
        inspectors = self.inspectors() if callable(self.inspectors) else self.inspectors

        for ready in asyncio.as_completed(
            [inspector.get_context(document, auto) for inspector in inspectors]
        ):
            if result := (await ready):
                return result
        return {}



"""Defines application settings."""

from __future__ import annotations

import json

from pygments.styles import STYLE_MAP as pygments_styles
from upath import UPath

from euporie.core.config import add_setting
from euporie.core.filters import in_mplex

# euporie.core.app.launch


def _euporie_core_launch_app_choices() -> list[str]:
    """List known euporie-apps and their aliases."""
    from euporie.core.__main__ import available_apps
    from euporie.core.app import APP_ALIASES

    return sorted(
        {entry.name for entry in available_apps().values()} - {"launch"}
        | APP_ALIASES.keys()
    )


add_setting(
    name="app",
    group="euporie.core.app.launch",
    flags=["app"],
    type_=str,
    help_="The application to launch",
    choices=_euporie_core_launch_app_choices,
    description="""
        The name of the application to launch.
    """,
)

# euporie.core.app

add_setting(
    name="files",
    group="euporie.core.app.app",
    default=[],
    flags=["files"],
    nargs="*",
    type_=UPath,
    help_="List of file names to open",
    schema={
        "type": "array",
        "items": {
            "description": "File path",
            "type": "string",
        },
    },
    description="""
        A list of file paths to open when euporie is launched.
    """,
)

add_setting(
    name="edit_mode",
    group="euporie.core.app.app",
    flags=["--edit-mode"],
    type_=str,
    choices=["micro", "emacs", "vi"],
    title="Editor key bindings",
    help_="Key-binding mode for text editing",
    default="micro",
    description="""
        Key binding style to use when editing cells.
    """,
)

add_setting(
    name="tab_size",
    group="euporie.core.app.app",
    flags=["--tab-size"],
    type_=int,
    help_="Spaces per indentation level",
    default=4,
    schema={
        "minimum": 1,
    },
    description="""
        The number of spaces to use per indentation level. Should be set to 4.
    """,
)

add_setting(
    name="terminal_polling_interval",
    group="euporie.core.app.app",
    flags=["--terminal-polling-interval"],
    type_=float,
    help_="Time between terminal colour queries",
    default=0.0,
    schema={
        "min": 0.0,
    },
    description="""
        Determine how frequently the terminal should be polled for changes to the
        background / foreground colours. Set to zero to disable terminal polling.
    """,
)

add_setting(
    name="formatters",
    group="euporie.core.app.app",
    flags=["--formatters"],
    type_=json.loads,
    help_="List of external code formatters",
    default=[
        # {"command": ["ruff", "format", "-"], "languages": ["python"]},
        # {"command": ["black", "-"], "languages": ["python"]},
        # {"command": ["isort", "-"], "languages": ["python"]},
    ],
    action="append",
    schema={
        "type": "array",
        "items": {
            "type": "object",
            "properties": {
                "command": {
                    "type": "array",
                    "items": [{"type": "string"}],
                },
                "languages": {
                    "type": "array",
                    "items": [{"type": "string", "unique": True}],
                },
            },
            "required": ["command", "languages"],
        },
    },
    description="""
        An array listing languages and commands of formatters to use for
        reformatting code cells. The command is an array of the command any any
        arguments. Code to be formatted is pass in via the standard input, and
        replaced with the standard output.

        e.g.

        .. code-block:: json

           [
             {"command": ["ruff", "format", "-"], "languages": ["python"]},
             {"command": ["black", "-"], "languages": ["python"]},
             {"command": ["isort", "-"], "languages": ["python"]}
           ]
    """,
)

add_setting(
    name="syntax_highlighting",
    group="euporie.core.app.base",
    flags=["--syntax-highlighting"],
    type_=bool,
    help_="Syntax highlighting",
    default=True,
    description="""
        Enable or disable syntax highlighting in code input fields.
    """,
)

add_setting(
    name="syntax_theme",
    group="euporie.core.app.base",
    flags=["--syntax-theme"],
    type_=str,
    help_="Syntax highlighting theme",
    default="euporie",
    schema={
        # Do not want to print all theme names in `--help` screen as it looks messy
        # so we only add them in the scheme, not as setting choices
        "enum": ["euporie", *pygments_styles.keys()],
    },
    description="""
        The name of the pygments style to use for syntax highlighting.
    """,
)

add_setting(
    name="color_depth",
    group="euporie.core.app.app",
    flags=["--color-depth"],
    type_=int,
    choices=[1, 4, 8, 24],
    default=None,
    help_="The color depth to use",
    description="""
        The number of bits to use to represent colors displayable on the screen.
        If set to None, the supported color depth of the terminal will be detected
        automatically.
    """,
)

add_setting(
    name="multiplexer_passthrough",
    group="euporie.core.app.app",
    flags=["--multiplexer-passthrough"],
    type_=bool,
    help_="Use passthrough from within terminal multiplexers",
    default=False,
    hidden=~in_mplex,
    description="""
        If set and euporie is running inside a terminal multiplexer
        (:program:`screen` or :program:`tmux`), then certain escape sequences
        will be passed-through the multiplexer directly to the terminal.

        This affects things such as terminal color detection and graphics display.

        for tmux, you will also need to ensure that ``allow-passthrough`` is set to
        ``on`` in your :program:`tmux` configuration.

        .. warning::

           Terminal graphics in :program:`tmux` is experimental, and is not
           guaranteed to work. Use at your own risk!

        .. note::
           As of version :command:`tmux` version ``3.4`` sixel graphics are
           supported, which may result in better terminal graphics then using
           multiplexer passthrough.
    """,
)

add_setting(
    name="color_scheme",
    group="euporie.core.app.app",
    flags=["--color-scheme"],
    type_=str,
    choices=["default", "inverse", "light", "dark", "black", "white", "custom"],
    help_="The color scheme to use",
    default="default",
    description="""
        The color scheme to use: `auto` means euporie will try to use your
        terminal's color scheme, `light` means black text on a white background,
        and `dark` means white text on a black background.
    """,
)

add_setting(
    name="custom_background_color",
    group="euporie.core.app.app",
    flags=["--custom-background-color", "--custom-bg-color", "--bg"],
    type_=str,
    help_='Background color for "Custom" color theme',
    default="#073642",
    schema={
        "maxLength": 7,
    },
    description="""
        The hex code of the color to use for the background in the "Custom" color
        scheme.
    """,
)

add_setting(
    name="custom_foreground_color",
    group="euporie.core.app.app",
    flags=["--custom-foreground-color", "--custom-fg-color", "--fg"],
    type_=str,
    help_='Foreground color for "Custom" color theme',
    default="#839496",
    schema={
        "maxLength": 7,
    },
    description="""
        The hex code of the color to use for the foreground in the "Custom" color
        scheme.
    """,
)

add_setting(
    name="accent_color",
    group="euporie.core.app.app",
    flags=["--accent-color"],
    type_=str,
    help_="Accent color to use in the app",
    default="ansiblue",
    description="""
        The hex code of a color to use for the accent color in the application.
    """,
)

add_setting(
    name="custom_styles",
    group="euporie.core.style",
    flags=["--custom-styles"],
    type_=json.loads,
    default={},
    schema={
        "type": "object",
    },
    help_="Additional style settings",
    description="""
        A JSON object mapping style names to prompt-toolkit style values.

        The style keys used in euporie can be found in :py:func:`euporie.core.style.build_style`.

        e.g.:

        .. code-block:: json

           { "cell input prompt":"fg:purple", "cell output prompt": "fg:green" }

    """,
)

add_setting(
    name="key_bindings",
    group="euporie.core.app.app",
    flags=["--key-bindings"],
    type_=json.loads,
    help_="Additional key binding definitions",
    default={},
    schema={
        "type": "object",
    },
    description="""
        A mapping of component names to mappings of command name to key-binding lists.
""",
)

add_setting(
    name="graphics",
    group="euporie.core.app.app",
    flags=["--graphics"],
    choices=["none", "sixel", "kitty", "kitty-unicode", "iterm"],
    type_=str,
    default=None,
    help_="The preferred graphics protocol",
    description="""
        The graphics protocol to use, if supported by the terminal.
        If set to "none", terminal graphics will not be used.
""",
)

add_setting(
    name="force_graphics",
    group="euporie.core.app.app",
    flags=["--force-graphics"],
    type_=bool,
    default=False,
    help_="Force use of specified graphics protocol",
    description="""
        When set to :py:const:`True`, the graphics protocol specified by the
        :option:`graphics` configuration option will be used even if the terminal
        does not support it.

        This is also useful if you want to use graphics in :command:`euporie-hub`.
""",
)

add_setting(
    name="enable_language_servers",
    group="euporie.core.app.app",
    flags=["--enable-language-servers", "--lsp"],
    menu_title="Language servers",
    type_=bool,
    default=False,
    help_="Enable language server support",
    description="""
        When set to :py:const:`True`, language servers will be used for liniting,
        code inspection, and code formatting.

        Additional language servers can be added using the
        :option:`language-servers` option.
""",
)

add_setting(
    name="language_servers",
    group="euporie.core.app.app",
    flags=["--language-servers"],
    type_=json.loads,
    help_="Language server configurations",
    default={},
    schema={
        "type": "object",
        "items": {
            "type": "object",
            "patternProperties": {
                "^[0-9]+$": {
                    "type": "object",
                    "properties": {
                        "command": {
                            "type": "array",
                            "items": [{"type": "string"}],
                        },
                        "language": {
                            "type": "array",
                            "items": [{"type": "string", "unique": True}],
                        },
                    },
                    "required": ["command"],
                }
            },
        },
    },
    description="""
        Additional language servers can be defined here, e.g.:

        .. code-block:: json

           {
            "ruff": {"command": ["ruff-lsp"], "languages": ["python"]},
            "pylsp": {"command": ["pylsp"], "languages": ["python"]},
            "typos": {"command": ["typos-lsp"], "languages": []}
           }

        The following properties are required:
        - The name to be given to the the language server, must be unique
        - The command list consists of the process to launch, followed by any
          command line arguments
        - A list of language the language server supports. If no languages are
        given, the language server will be used for documents of any language.

        To disable one of the default language servers, its name can be set to an
        empty dictionary. For example, the following would disable the awk language
        server:

        .. code-block:: json

           {
             "awk-language-server": {},
           }
    """,
)

# euporie.core.app.cursor

add_setting(
    name="set_cursor_shape",
    group="euporie.core.app.cursor",
    flags=["--set-cursor-shape"],
    type_=bool,
    default=True,
    menu_title="Change cursor shape",
    help_="Whether to set the shape of the cursor depending on the editing mode",
    description="""
        When set to True, the euporie will set the shape of the terminal's cursor
        to a beam in insert mode and and underline in replace mode when editing.
""",
)

add_setting(
    name="cursor_blink",
    group="euporie.core.app.cursor",
    flags=["--cursor-blink"],
    type_=bool,
    default=False,
    help_="Whether to blink the cursor",
    description="""
        When set to True, the cursor will blink.
""",
)



"""Defines a web-view tab."""

from __future__ import annotations

import asyncio
import logging
from typing import TYPE_CHECKING

from prompt_toolkit.filters import Condition
from prompt_toolkit.layout.containers import HSplit, VSplit
from prompt_toolkit.layout.dimension import Dimension
from upath import UPath

from euporie.core.app.current import get_app
from euporie.core.convert.mime import get_mime
from euporie.core.data_structures import DiBool
from euporie.core.layout.decor import FocusedStyle
from euporie.core.margins import MarginContainer, ScrollbarMargin
from euporie.core.tabs.base import Tab
from euporie.core.widgets.display import DisplayWindow
from euporie.core.widgets.forms import Button, Text
from euporie.web.widgets.webview import WebViewControl

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Any, Callable, ClassVar

    from prompt_toolkit.layout.containers import AnyContainer

    from euporie.core.app.app import BaseApp
    from euporie.core.bars.status import StatusBarFields

log = logging.getLogger(__name__)


class WebTab(Tab):
    """Tab class for displaying files."""

    name = "Web Viewer"
    weight = 2
    mime_types: ClassVar[set[str]] = {"text/html", "text/markdown"}

    def __init__(self, app: BaseApp, path: Path | None) -> None:
        """Call when the tab is created."""
        super().__init__(app, path)
        self.status: Callable[[], StatusBarFields] | None = None

        def _load() -> None:
            old_container = self.container
            self.container = self.load_container()
            if self.app.layout.has_focus(old_container):
                self.focus()
            self.app.invalidate()

        self._load_task = app.create_background_task(asyncio.to_thread(_load))

    def focus(self) -> None:
        """Focus the webview when this tab is focused."""
        self.app.layout.focus(self.webview)

    @property
    def title(self) -> str:
        """Return the tab title."""
        if (webview := getattr(self, "webview", None)) and (title := webview.title):
            return title
        if self.path is not None:
            return str(self.path.name) or str(self.path)
        else:
            return "<Web Page>"

    def load_url(self, url: str | Path, new_tab: bool = False, **kwargs: Any) -> bool:
        """Load a new URL, or the URL in the address-bar."""
        log.debug("Loading %s", url)
        self.webview.loading = True
        self.app.invalidate()
        if not url:
            return False
        if isinstance(url, str):
            url = UPath(url, protocol="https" if ":" not in url else None)
        if not new_tab and get_mime(url) in self.mime_types:
            self.webview.load_url(url, **kwargs)
        else:
            self.webview.loading = False
            get_app().open_file(url)
        return True

    def _url_loaded(self, webview: WebViewControl) -> None:
        """Trigger callback when the URL is loaded."""
        url = webview.url
        if url is not None:
            self.path = UPath(url)
            self.url_bar.text = str(url)
        # Focus the webview
        self.app.layout.current_control = self.webview

    def load_container(self) -> AnyContainer:
        """Abcract method for loading the notebook's main container."""
        assert self.path is not None
        path = self.path
        self.webview = WebViewControl(url=path, link_handler=self.load_url)
        self.webview.rendered += self._url_loaded

        def _status() -> StatusBarFields:
            return (self.webview.status if self.webview is not None else [], [])

        self.status = _status

        button_prev = Button(
            "◀",
            show_borders=DiBool(top=True, right=False, bottom=True, left=True),
            disabled=Condition(lambda: not self.webview.prev_stack[:-1]),
            on_click=lambda x: self.webview.nav_prev(),
        )
        button_next = Button(
            "▶",
            show_borders=DiBool(top=True, right=True, bottom=True, left=False),
            disabled=Condition(lambda: not self.webview.next_stack),
            on_click=lambda x: self.webview.nav_next(),
        )
        button_refresh = Button(
            "↻",
            on_click=lambda x: (self.load_url(self.path) and False) or None,
        )

        def _select_url() -> None:
            """Select all in url bar when it gains focus."""
            from prompt_toolkit.selection import SelectionState

            buffer = self.url_bar.buffer
            buffer.selection_state = SelectionState(0)
            buffer.cursor_position = len(buffer.text)
            buffer.selection_state.enter_shift_mode()

        self.url_bar = Text(
            text=str(path),
            show_borders=DiBool(top=True, right=False, bottom=True, left=True),
            accept_handler=lambda buf: self.load_url(buf.text),
            on_focus=_select_url,
        )
        button_go = Button(
            "➜",
            show_borders=DiBool(top=True, right=True, bottom=True, left=False),
            on_click=lambda x: (self.load_url(self.url_bar.text) and None)
            or None,  # Magical typing magic
        )

        return HSplit(
            [
                VSplit(
                    [
                        FocusedStyle(button_prev),
                        FocusedStyle(button_next),
                        FocusedStyle(button_refresh),
                        FocusedStyle(self.url_bar),
                        FocusedStyle(button_go),
                    ],
                    style="class:tab.toolbar",
                ),
                VSplit(
                    [
                        window := DisplayWindow(self.webview),
                        MarginContainer(ScrollbarMargin(), target=window),
                    ]
                ),
            ],
            width=Dimension(weight=1),
            height=Dimension(weight=1),
        )

    def __pt_status__(self) -> StatusBarFields:
        """Return a list of statusbar field values shown then this tab is active."""
        if callable(self.status):
            return self.status()
        else:
            return ([], [])



"""Define settings for the preview app."""

from __future__ import annotations

import logging

from upath import UPath

from euporie.core.config import add_setting

log = logging.getLogger(__name__)


add_setting(
    name="output_file",
    group="euporie.preview.app",
    flags=["--output-file"],
    nargs="?",
    default="-",
    const="-",
    type_=UPath,
    help_="Output path when previewing file",
    description="""
        When set to a file path, the formatted output will be written to the
        given path. If no value is given (or the default "-" is passed) output
        will be printed to standard output.
    """,
)

add_setting(
    name="page",
    group="euporie.preview.app",
    flags=["--page"],
    type_=bool,
    help_="Pass output to pager",
    default=False,
    description="""
        Whether to pipe output to the system pager when previewing a notebook.
    """,
)


add_setting(
    name="run",
    group="euporie.preview.app",
    flags=["--run"],
    type_=bool,
    help_="Run the notebook files when loaded",
    default=False,
    description="""
        If set, notebooks will be run automatically when opened, or if previewing a
        file, the notebooks will be run before being output.
    """,
)

add_setting(
    name="save",
    group="euporie.preview.app",
    flags=["--save"],
    type_=bool,
    help_="Save the notebook after running it",
    default=False,
    description="""
        If set, notebooks will be saved after they have been run. This setting only
        has any affect if the :option:`run` setting is active.
    """,
)

add_setting(
    name="show_filenames",
    group="euporie.preview.app",
    flags=["--show-filenames"],
    type_=bool,
    help_="Show the notebook filenames when previewing multiple notebooks",
    default=False,
    description="""
        If set, the notebook filenames will be printed above each notebook's output
        when multiple notebooks are being previewed.
    """,
)

add_setting(
    name="cell_start",
    group="euporie.preview.app",
    flags=["--cell-start"],
    type_=int,
    help_="The first cell to include in the preview",
    default=None,
    description="""
        When set, only cells after the given cell index will be shown.
    """,
)

add_setting(
    name="cell_stop",
    group="euporie.preview.app",
    flags=["--cell-stop"],
    type_=int,
    help_="The last cell to include in the preview",
    default=None,
    description="""
        When set, only cells before the given cell index will be shown.
    """,
)
